#include "lhwrbi.h"
#include "hint.h"
#include "bigarray.h"
#include "internalbigarray.h"
#include "btreecube.h"
#include "btreecube_rotated.h"
#include "cdbcube.h"
#include "psqlcube.h"
#include "fileutil.h"
#include <time.h>
#include <cmath>
#include <stdlib.h>
#include <vector>
#include <list>
#include <iostream>
#include <stack>

#include <climits>
#include <ctime>
#include <sys/resource.h>
#include <sys/time.h>

inline double diff(rusage start,rusage end) {
  return ((end.ru_utime.tv_sec+end.ru_utime.tv_usec*1e-6 +
	   end.ru_stime.tv_sec+end.ru_stime.tv_usec*1e-6) -
	  (start.ru_utime.tv_sec+start.ru_utime.tv_usec*1e-6 +
	   start.ru_stime.tv_sec+start.ru_stime.tv_usec*1e-6));
}

inline double diff(timeval start, timeval end){
  double start_time = start.tv_sec + (start.tv_usec * 1e-6);
  double end_time = end.tv_sec + (end.tv_usec * 1e-6);
  return (end_time - start_time);

}

void tableHeader(){
    cout << "\\begin{tabular}{crcccccr}" << endl;
    cout << "\\hline\\hline " << endl;
    cout << "Cube & N & put & get & slice & dice & rollup & size\\\\ \\hline" << endl;

}

void tableHeader(int numDims){
    cout << "\\begin{tabular}{cr";
    for(int i = 0; i < numDims; i++){
        cout << "c";
    }
    cout << "}" << endl;
    cout << "\\hline\\hline " << endl;
    cout << "Cube & N ";
    for(int i = 0; i < numDims; i++){
        cout << " & " << i ;
    }
    cout << "\\\\" << endl << "\\hline" << endl;
    

}

void tableFooter(){
    cout << "\\hline"<< endl;
    cout << "\\end{tabular}" << endl;
}
	
string latex_format(double d){
  int exponent = 0;
  while(d >= 10){
    d/=10;
    exponent++;
  }
  while(d < 1){
    d*=10;
    exponent--;
  }

  stringstream str;
  str << " $ ";
  d *= 100;
  int i = (int)floor(d+0.5);
  if(i == 1000){
    exponent ++;
  }
  stack<int> s;
  while(i > 0){
    s.push(i%10);
    (i /= 10);
  }
  str << s.top() << ".";
  s.pop();
  str << s.top();
  s.pop();
  str << s.top();
  s.pop();

  str << " \\times 10^{";
  if(exponent < 0){
    str << exponent;
  }
  else{
    str << "~" << exponent;
  }
  str << "} $";
  return str.str();

}

string latex_format(int i){
  if(i <= 0) return ("");
  stack<int> s;
  while(i > 0){
    s.push(i%10);
    i /= 10;
  }
  
  stringstream str;
  str << "$ ";
  do{
    str << s.top();
    s.pop();
    if(s.size() % 3 == 0 && s.size() > 0) str << ",";
  }while(s.size() > 0);
  str << " $";
  return str.str();
}


void test(MultidimensionalArray<int>& cube, string file, int wallclock = 1, int bs=100, int nhb=15){
  
  srand(0); //make all the tests as close to the same as possible
  int numDims = (int)cube.getShape().size();
  vector<int>  vals(0);
  vector<int> pos(numDims);
  double seconds;
  
  struct rusage start, finish;
  struct timeval t_start, t_finish, st,fi;

  cout << file << " & ";

    fstream fs(file.c_str());
    while(fs  && fs.good()){
        int temp;
        fs >> temp;	
        if(fs.good()){
            vals.push_back(temp);
        }
    }
    fs.close();

    int counter = 0;
    
    getrusage(RUSAGE_SELF, &start);
    gettimeofday(&t_start, (struct timezone*)0);
    for(uint i = 0; i < vals.size()/(numDims+1); i++){
        int offset = counter * (numDims+1);
	for(int j = 0; j < numDims; j++){
	    pos[j] = vals[offset + j];
	}
        int v = vals[offset + numDims];
	cube.put(pos,v);
	counter++;
    }
    getrusage(RUSAGE_SELF, &finish);
    gettimeofday(&t_finish, (struct timezone*)0);
    
    seconds =  diff(t_start,t_finish);    
    cout << latex_format(cube.getNumberOfRecords()) << " &";
    cout << latex_format(seconds/(double)counter) << " &";
    cout.flush();
    getrusage(RUSAGE_SELF, &start);
    gettimeofday(&t_start, (struct timezone*)0);

    int count = 0;
    for(uint i = 0; i < vals.size()/(numDims+1); i++){
        int offset = count * (numDims+1);
	for(int j = 0; j < numDims; j++){
	    pos[j] = vals[offset + j];
	}
	cube.get(pos);
	count++;
    }
    getrusage(RUSAGE_SELF, &finish);
    gettimeofday(&t_finish, (struct timezone*)0);
    if(wallclock)
      seconds =  diff(t_start,t_finish);
    else
      seconds =  diff(start,finish);
    cout << latex_format(seconds/(double)count) << " & ";
    cout.flush();


    Hint h;
    h.setBucketSize(bs);
    h.setNumHashingBits(nhb);
    if(numDims > 1){
        getrusage(RUSAGE_SELF, &start);
	gettimeofday(&t_start, (struct timezone*)0);
        count = 0;
        while(count == 0 || diff(start,finish) < 0.5){
            for(int i = 0; i < (int)cube.getShape().size(); i++){
                int r = (int)floor((double)rand()/RAND_MAX*(int)vals.size()/(numDims+1));
                MultidimensionalArray<int>* sliced= cube.slice(i,vals[r*(numDims+1) + i] ,"slicecube",0,h);
                cerr << "Slicing dim " << i << " on " << vals[r*(numDims+1) + i] << " resulting in a cube of size " << sliced->getNumberOfRecords() << endl;
                cerr << "distribution value = " << sliced->maxInBucket() << endl;
                gettimeofday(&st, (struct timezone*)0);
                sliced->deleteOnDisk();
                delete sliced;
                gettimeofday(&fi, (struct timezone*)0);
                cerr << "Took " << diff(st,fi) << " to delete" << endl;
            }
            getrusage(RUSAGE_SELF, &finish);
	    gettimeofday(&t_finish, (struct timezone*)0);
            count ++;
        }
	if(wallclock)
	  seconds =  diff(t_start,t_finish);
	else
	  seconds =  diff(start,finish);

        cout << latex_format(seconds/(double)numDims/(double)count) << " & ";
    }
    else{
        cout << " & ";
    }
    cout.flush();

    //Try to keep as many of the random numbers the same as possible
    for(int i = count*numDims; i < 100*numDims; i++){
        rand();
    }

    //    h.setBucketSize((int)(bs));
    getrusage(RUSAGE_SELF, &start);
    gettimeofday(&t_start, (struct timezone*)0);
    count = 0;
    while(count == 0 || diff(start,finish) < 0.5){
        for(int i = 0; i < (int)cube.getShape().size(); i++){
	    int r1 = (int)floor((double)rand()/RAND_MAX*(int)vals.size()/(numDims+1));
	    int r2 = (int)floor((double)rand()/RAND_MAX*(int)vals.size()/(numDims+1));
            InvertedAgMap mymap[numDims];
            int a = min(vals[r1*(numDims+1) + i],vals[r2*(numDims+1) + i]); 
            int b = max(vals[r1*(numDims+1) + i],vals[r2*(numDims+1) + i]); 
            int dim = i;
	    //	    cerr << "a = " << a << ", b = " << b << endl;

            mymap[dim].setNumberOfAttributeValues(b+1);
            for(int j = a; j <= b; j++){
                vector<int> ancestors(1);
                vector<double> weights(1);
                ancestors[0] = j-a;
                weights[0] = 1;
                mymap[dim].setAttributeValueAncestors(j,ancestors,weights);
            }
            MultidimensionalArray<int>* diced= cube.general_dice(mymap,"dicecube",0,h);  
            diced->deleteOnDisk();
	    delete diced;
	}
	getrusage(RUSAGE_SELF, &finish);
	gettimeofday(&t_finish, (struct timezone*)0);
	count ++;
    }
    if(wallclock)
      seconds =  diff(t_start,t_finish);
    else
      seconds =  diff(start,finish);

    //Try to keep as many of the random numbers the same as possible
    for(int i = count*numDims*2; i < 100 * numDims; i++){
        rand();
    }

    cout << latex_format(seconds/(double)numDims/(double)count) << " & ";
    cout.flush();

    //    h.setBucketSize((int)(bs * (double)(numDims-1)/(double)numDims));
    if(numDims > 1){
        getrusage(RUSAGE_SELF, &start);
	gettimeofday(&t_start, (struct timezone*)0);
        count = 0;
        while(count == 0 || diff(start,finish) < 0.5){
            for(int i = 0; i < (int)cube.getShape().size(); i++){
                vector<int> killDims(1);
                killDims[0] = i;
                MultidimensionalArray<int>* rollup = cube.full_rollup(killDims, "rollupcube",0,h);
                cerr << "Rolling up dim " << i << " resulting in a cube of size " << rollup->getNumberOfRecords() << endl;
                cerr << "distribution value = " << rollup->maxInBucket() << endl;
                gettimeofday(&st, (struct timezone*)0);
                rollup->deleteOnDisk();
                delete rollup;
                gettimeofday(&fi, (struct timezone*)0);
                cerr << "Took " << diff(st,fi) << " to delete" <<endl;
                
            }
            getrusage(RUSAGE_SELF, &finish);
	    gettimeofday(&t_finish, (struct timezone*)0);
            count ++;
        }
	if(wallclock)
	  seconds =  diff(t_start,t_finish);
	else
	  seconds =  diff(start,finish);
        cout << latex_format(seconds/(double)numDims/(double)count) << " & ";
	
    }
    else{
      cout << " & ";
    }
    cout.flush();
    
    cout << latex_format((int)cube.getSizeOnDisk()) << "\\\\" << endl;    
    
}


void psqltest(PostgreSQLCube<int>& cube, string file, int bs=100){
  
  srand(0); //make all the tests as close to the same as possible
  int numDims = (int)cube.getShape().size();
  vector<int>  vals(0);
  vector<int> pos(numDims);
  double seconds;
  int count = 0;
  Hint h;
  
  struct rusage start, finish;
  struct timeval t_start, t_finish;
  cout << file << " & ";
  
  fstream fs(file.c_str());
  while(fs  && fs.good()){
    int temp;
    fs >> temp;	
    if(fs.good()){
      vals.push_back(temp);
    }
  }
  fs.close();
  //    cerr << "Done reading the file..." << endl;
    int counter = 0;
      
    
    getrusage(RUSAGE_SELF, &start);
    gettimeofday(&t_start, (struct timezone*)0);
    cube.setLoading(1);
    for(uint i = 0; i < vals.size()/(numDims+1); i++){
        int offset = counter * (numDims+1);
	for(int j = 0; j < numDims; j++){
	    pos[j] = vals[offset + j];
	}
        int v = vals[offset + numDims];
	cube.put(pos,v);
	counter++;
    }
    cube.setLoading(0);
    getrusage(RUSAGE_SELF, &finish);
    gettimeofday(&t_finish, (struct timezone*)0);
    seconds =  diff(t_start,t_finish);    
    cout << latex_format(cube.getNumberOfRecords()) << " & ";
    cout << latex_format(seconds/(double)counter) << " & ";
    cout.flush();
    seconds = 0;
    getrusage(RUSAGE_SELF, &start);
    gettimeofday(&t_start, (struct timezone*)0);
    count = 0;
    for(uint i = 0; i < vals.size()/(numDims+1) && seconds < 5; i++){
        int offset = count * (numDims+1);
	for(int j = 0; j < numDims; j++){
	    pos[j] = vals[offset + j];
	}
	cube.get(pos);
	count++;
	if(count % 1000 == 0){
	  getrusage(RUSAGE_SELF, &finish);
	  gettimeofday(&t_finish, (struct timezone*)0);
	  seconds =  diff(t_start,t_finish);
	}   
    }
    getrusage(RUSAGE_SELF, &finish);
    gettimeofday(&t_finish, (struct timezone*)0);
    seconds =  diff(t_start,t_finish);    
    cout << latex_format(seconds/(double)count) << " & ";
    cout.flush();

    h.setBucketSize((int)(bs * (double)(numDims-1)/(double)numDims));
    if(numDims > 1){
        getrusage(RUSAGE_SELF, &start);
	gettimeofday(&t_start, (struct timezone*)0);
        count = 0;
        while(count == 0 || diff(start,finish) < 0.5){
            for(int i = 0; i < (int)cube.getShape().size(); i++){
                int r = (int)floor((double)rand()/RAND_MAX*(int)vals.size()/(numDims+1));
                MultidimensionalArray<int>* sliced= cube.slice(i,vals[r*(numDims+1) + i] ,"slicecube",0,h);
                sliced->deleteOnDisk();
                delete sliced;
            }
            getrusage(RUSAGE_SELF, &finish);
	    gettimeofday(&t_finish, (struct timezone*)0);
            count ++;
        }
        seconds =  diff(t_start,t_finish);
        cout << latex_format(seconds/(double)numDims/(double)count) << " & ";
      
    }
    else{
        cout << " & ";
    }
    cout.flush();

    //Try to keep as many of the random numbers the same as possible
    for(int i = count*numDims; i < 100*numDims; i++){
        rand();
    }
  
    h.setBucketSize((int)(bs));
    getrusage(RUSAGE_SELF, &start);
    gettimeofday(&t_start, (struct timezone*)0);
    count = 0;
    while(count == 0 || diff(start,finish) < 0.5){
        for(int i = 0; i < (int)cube.getShape().size(); i++){
	    int r1 = (int)floor((double)rand()/RAND_MAX*(int)vals.size()/(numDims+1));
	    int r2 = (int)floor((double)rand()/RAND_MAX*(int)vals.size()/(numDims+1));
            InvertedAgMap mymap[numDims];
            int a = min(vals[r1*(numDims+1) + i],vals[r2*(numDims+1) + i]); 
            int b = max(vals[r1*(numDims+1) + i],vals[r2*(numDims+1) + i]); 
            int dim = i;
	    //	    cerr << "a = " << a << ", b = " << b << endl;
            mymap[dim].setNumberOfAttributeValues(b+1);
            for(int j = a; j <= b; j++){
	      //	        cout << j << endl;
                vector<int> ancestors(1);
                vector<double> weights(1);
                ancestors[0] = j-a;
                weights[0] = 1.0;
                mymap[dim].setAttributeValueAncestors(j,ancestors,weights);
            }
            MultidimensionalArray<int>* diced= cube.general_dice(mymap,"dicecube",0,h);  
            diced->deleteOnDisk();
	    delete diced;
	}
	getrusage(RUSAGE_SELF, &finish);
	gettimeofday(&t_finish, (struct timezone*)0);
	count ++;
    }
    seconds =  diff(t_start,t_finish);

    //Try to keep as many of the random numbers the same as possible
    for(int i = count*numDims*2; i < 100 * numDims; i++){
        rand();
    }

    cout << latex_format(seconds/(double)numDims/(double)count) << " & ";
    cout.flush();

    h.setBucketSize((int)(bs * (double)(numDims-1)/(double)numDims));
    if(numDims > 1){
        getrusage(RUSAGE_SELF, &start);
	gettimeofday(&t_start, (struct timezone*)0);
        count = 0;
        while(count == 0 || diff(start,finish) < 0.5){
            for(int i = 0; i < (int)cube.getShape().size(); i++){
                vector<int> killDims(1);
                killDims[0] = i;
                MultidimensionalArray<int>* rollup = cube.full_rollup(killDims, "rollupcube",0,h);
                rollup->deleteOnDisk();
                delete rollup;
                
            }
            getrusage(RUSAGE_SELF, &finish);
	    gettimeofday(&t_finish, (struct timezone*)0);
            count ++;
        }
        seconds =  diff(t_start,t_finish);
        cout << latex_format(seconds/(double)numDims/(double)count) << " & ";
    }
    else{
        cout << " & ";
    }
    cout << latex_format((int)cube.getSizeOnDisk()) << "\\\\" << endl;    

}


void slicetest(MultidimensionalArray<int>& cube, string file, int bs = 100,int wallclock = 1){
  
  srand(0); //make all the tests as close to the same as possible
  int numDims = (int)cube.getShape().size();
  vector<int>  vals(0);
  vector<int> pos(numDims);
  double seconds;
  
  struct rusage start, finish;
  struct timeval t_start, t_finish;

  cout << file;
  cout.flush();

    fstream fs(file.c_str());
    while(fs  && fs.good()){
        int temp;
        fs >> temp;	
        if(fs.good()){
            vals.push_back(temp);
        }
    }
    fs.close();

    int counter = 0;

   
    for(uint i = 0; i < vals.size()/(numDims+1); i++){
        int offset = counter * (numDims+1);
	for(int j = 0; j < numDims; j++){
	    pos[j] = vals[offset + j];
	}
        int v = vals[offset + numDims];
	cube.put(pos,v);
	counter++;
    }

    Hint h;
    h.setBucketSize((int)(bs));

    int count;
    for(int i = 0; i < (int)cube.getShape().size(); i++){
        count = 0;
        getrusage(RUSAGE_SELF, &start);
	gettimeofday(&t_start, (struct timezone*)0);
        while(count == 0 || diff(start,finish) < 0.5){
	    int r = (int)floor((double)rand()/RAND_MAX*(int)vals.size()/(numDims+1));
	    MultidimensionalArray<int>* sliced= cube.slice(i,vals[r*(numDims+1) + i] ,"slicecube",0,h);
	    sliced->deleteOnDisk();
	    delete sliced;
            getrusage(RUSAGE_SELF, &finish); 
	    gettimeofday(&t_finish, (struct timezone*)0);
	    count ++;
	}
	if(wallclock){
	  seconds =  diff(t_start,t_finish);
	}
	else{
	  seconds =  diff(start,finish);
	}
        cout << " & " << latex_format(seconds/(double)count);
	cout.flush();

    }   

    cout << "\\\\" << endl;
}
void psqlslicetest(PostgreSQLCube<int>& cube, string file, int bs = 100){
  
  srand(0); //make all the tests as close to the same as possible
  int numDims = (int)cube.getShape().size();
  vector<int>  vals(0);
  vector<int> pos(numDims);
  double seconds;
  
  struct rusage start, finish;
  struct timeval t_start, t_finish;
  cout << file;
  cout.flush();

    fstream fs(file.c_str());
    while(fs  && fs.good()){
        int temp;
        fs >> temp;	
        if(fs.good()){
            vals.push_back(temp);
        }
    }
    fs.close();

    int counter = 0;

    cube.setLoading(1);
    for(uint i = 0; i < vals.size()/(numDims+1); i++){
        int offset = counter * (numDims+1);
	for(int j = 0; j < numDims; j++){
	    pos[j] = vals[offset + j];
	}
        int v = vals[offset + numDims];
	cube.put(pos,v);
	counter++;
    }
    cube.setLoading(0);

    Hint h;
    h.setBucketSize((int)(bs));

    int count;
    for(int i = 0; i < (int)cube.getShape().size(); i++){
        count = 0;
        getrusage(RUSAGE_SELF, &start);
	gettimeofday(&t_start, (struct timezone*)0);
        while(count == 0 || diff(start,finish) < 0.5){
	    int r = (int)floor((double)rand()/RAND_MAX*(int)vals.size()/(numDims+1));
	    MultidimensionalArray<int>* sliced= cube.slice(i,vals[r*(numDims+1) + i] ,"slicecube",0,h);
	    sliced->deleteOnDisk();
	    delete sliced;
            getrusage(RUSAGE_SELF, &finish);
	    gettimeofday(&t_finish, (struct timezone*)0);
            count ++;
	}
        seconds =  diff(t_start,t_finish);
        cout << " & " << latex_format(seconds/(double)count);
    	cout.flush();

    }   

    cout << "\\\\" << endl;
}

	
int main() {
    
    vector<string> files(5);
    files[0] = "file1.out";
    files[1] = "file1.5.out";
    files[2] = "file2.out";
    files[3] = "file3.out";
    files[4] = "file4.out";
    vector<int> numDims(5);
    numDims[0] = 1;
    numDims[1] = 3;
    numDims[2] = 3;
    numDims[3] = 3;
    numDims[4] = 3;

    vector<vector<int> > shape(3);
    shape[0].push_back(20177);
    shape[1].push_back(18);
    shape[1].push_back(18);
    shape[1].push_back(6);
    shape[2].push_back(16);
    shape[2].push_back(16);
    shape[2].push_back(6);

    vector<int> bucketSize(5);
    bucketSize[0] = 100;
    bucketSize[1] = 100;
    bucketSize[2] = 100;
    bucketSize[3] = 100;//4000;
    bucketSize[4] = 100;//8000;
    
/*    cout << "\\textbf{OPERATION TESTS}\\\\" << endl;

    cout << "\\textbf{PostgreSQL Cube}\\\\" << endl;
    tableHeader();
    for(uint i = 0; i < files.size(); i++){
	PostgreSQLCube<int> psqlc(numDims[i],"postgrestest");
	psqltest(psqlc,files[i]);
	psqlc.deleteOnDisk();
    }
    tableFooter();
    cout << endl;
    

    cout << "\\textbf{Big Array Cube}\\\\" << endl;
    tableHeader();
    for(uint i = 0; i < 3; i++){
        BigArray<int> ba(shape[i],  "bigarray.bin");
	test(ba, files[i]);
	ba.deleteOnDisk();
    }
    tableFooter();
    cout << endl;    
*/
    int numBits = 15;
    int wallclock = 1;
	cout << "\\textbf{LHWRBI Cube}\\\\" << endl;
	if(wallclock) cout << "Wallclock timing" << endl;
	cout << "numBits = " << numBits << endl;
	tableHeader();
	
	//do this one later... don't forget!!
	
	for(uint i = 3; i < files.size(); i++){
	  LinearHashCube<int> lhc(numDims[i],"linearhashtest.bin", 100,numBits);
	  test(lhc,files[i],wallclock, 100,numBits);
	  lhc.deleteOnDisk();
	}
	
	tableFooter();
	cout << endl;
      
    

    
    cout << "\\textbf{B-Tree Cube}\\\\" << endl;
    tableHeader();
    for(uint i = 4; i < files.size(); i++){
	BTreeCube<int> btc(numDims[i],  "btreecube.bin");
	test(btc,files[i]);
	btc.deleteOnDisk();
    }
    tableFooter();
    cout << endl;      
    
    cout << "\\textbf{Rotated B-Tree Cube}\\\\" << endl;
    tableHeader();
    for(uint i = 4; i < files.size(); i++){
        set<int> rotations;
        for(int k = 0; k < numDims[i]; k++){
            rotations.insert(k);
        }
	BTreeCube_Rotated<int> btcr(numDims[i], rotations, "btreecuberotated.bin");
	test(btcr,files[i]);
	btcr.deleteOnDisk();
    }
    tableFooter();
    cout << endl;      
        
    
    //    cout << "\\textbf{SLICING TESTS}\\\\" << endl;
    /*
    cout << "\\textbf{PostgreSQL Cube}\\\\" << endl;
    tableHeader(3);
    for(uint i = 1; i < files.size(); i++){
	PostgreSQLCube<int> psqlc(numDims[i],  "postgresqlcube");
	psqlslicetest(psqlc,files[i]);
	psqlc.deleteOnDisk();
    }
    tableFooter();
    cout << endl;      

    cout << "\\textbf{Big Array Cube}\\\\" << endl;
    tableHeader(3);
    for(uint i = 1; i < 3; i++){
        BigArray<int> ba(shape[i],  "bigarray.bin");
	slicetest(ba, files[i]);
	ba.deleteOnDisk();
    }
    tableFooter();
    cout << endl;    
    */
    
    /*    for(int wallclock = 0; wallclock < 2; wallclock ++){
      cout << "\\textbf{LHWRBI Cube}\\\\" << endl;
      tableHeader(3);
      for(uint i = 1; i < files.size(); i++){
	LinearHashCube<int> lhc(numDims[i],"linearhashtest.bin", bucketSize[i],15);
	slicetest(lhc,files[i],bucketSize[i], wallclock);
	lhc.deleteOnDisk();
      }
      tableFooter();
      cout << endl;    
    }
    */
    /*
    cout << "\\textbf{B-Tree Cube}\\\\" << endl;
    tableHeader(3);
    for(uint i = 1; i < files.size(); i++){
	BTreeCube<int> btc(numDims[i],  "btreecube.bin");
	slicetest(btc,files[i]);
	btc.deleteOnDisk();
    }
    tableFooter();
    cout << endl;      

    cout << "\\textbf{Rotated B-Tree Cube}\\\\" << endl;
    tableHeader(3);
    for(uint i = 1; i < files.size(); i++){
        set<int> rotations;
        for(int k = 0; k < numDims[i]; k++){
            rotations.insert(k);
        }
	BTreeCube_Rotated<int> btcr(numDims[i], rotations, "btreecuberotated.bin");
	slicetest(btcr,files[i]);
	btcr.deleteOnDisk();
    }
    tableFooter();
    cout << endl;      
    */

}
 
