//Steve is responsible for this as well
#ifndef BUCKET_H
#define BUCKET_H

#include <vector>
#include <stdexcept>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <sys/mman.h>
#include "multidimensionalarray.h"
#include <string>
#include <fstream>
#include <cassert>
#include <sstream>
#include <iostream>
#include <stdlib.h>
#include <limits.h>
#include <stdio.h>
#include "fileutil.h"
#include <list>

using namespace std;

//template<class DataType>
//class BucketData; 
template<class DataType>
class BucketIterator; 

template<class DataType>
class Bucket{


    public:
            /*****
             * ALWAYS INITIALIZE all variables in C++, especially
             * the pointers. Having pointers that point to random
             * location in memory is insane!!!!!!!!!!
             */
    Bucket():mNumberOfDimensions(0), mBucketSize(0),mIndex(0), mDepth(0),
	/*data(0),*/ hash(0),/*numItems(0),*/next(0),
	mFileDescriptor(0),mData(0),mFileName(),
	mBaseFileName()
	{
//	    cerr<<"BUCKET Oh Oh, default constructor called! "<<endl;
	    mRefCount = new int(1);
	}
    
    Bucket(const Bucket<DataType> & other):
	mNumberOfDimensions(other.mNumberOfDimensions),
	mBucketSize(other.mBucketSize),
	mIndex(other.mIndex), 
	mDepth(other.mDepth),
	/*data(other.data), */
	hash(other.hash),
	/*numItems(other.numItems),*/
	next(other.next), 
	mFileDescriptor(other.mFileDescriptor),
	mData(other.mData),
	mFileName(other.mFileName), 
	mBaseFileName(other.mBaseFileName),
	mRefCount(other.mRefCount)
	{
//	    cerr << "BUCKET COPY CONSTRUCTOR!!! MIGHT BE INCORRECT!!!"<<endl;
	    (*mRefCount)++;
//	    cout << "RefCount now " << *mRefCount << endl;
	}
    
    /**
     * Always initialize variables using the notation below. It is
     * safer, faster.
     */
    Bucket(int numDimensions, int size, int h, int index, string file, int depth = 0): 
	mNumberOfDimensions(numDimensions), mBucketSize(size),mIndex(index), mDepth(depth),
	/*data(vector<BucketData<DataType> >(0)),*/ hash(h),/*numItems(0),*/next(0), 
	mFileDescriptor(0),mData(0),mFileName(), mBaseFileName()
	{
	    mRefCount = new int(1);
	    if(verbose)cout << "Bucket Constructor Called..." << this->mBucketSize << endl;
	    stringstream s;
	    s << file << "-" <<this->mIndex << "-" << this->mDepth;
	    this->mFileName = s.str();
	    this->mBaseFileName = file;
	    this->open();
	    
	}
    
    ~Bucket(){
	if(verbose)cout << "Reference Count = " << *mRefCount << endl;
	if((*mRefCount) == 1){
//	    cout << "mRefCount = " << *mRefCount << endl;
	    if(next != 0){
		delete next;
                next = 0;

	    }
	    this->close();
	    delete mRefCount;
            mRefCount = 0;
	    delete mData;
            mData = 0;
	}
	else{
	    if(verbose)cout << "Bucket Index = " <<this->mIndex << endl;
	    (*mRefCount) --;	    
	}
    }
    
    void getFileNames(list<string> * ret){
	ret->push_back(this->mFileName);
	if(next!=0){
	    next->getFileNames(ret);
	}
	return;
    }

    inline int getHash() const {return hash;}

    inline void setHash(int inHash){this->hash = inHash;}
    
    inline int getNumItems() const {
//	cout <<" getNumItems "<<endl;
//	cout << ((int)(mData[0])) << endl;
//	cout <<" getNumItems ok"<<endl;
	if(mData == 0) {cerr << "This will crash, you haven't called open yet! (default constructor?)"<<endl;}
	return *((int *)(mData));
    }
    
    inline int getSize() const {return mBucketSize;}

    inline Bucket* getNext() const{return next;}

    void open() throw(std::ifstream::failure);

    void close();

    void blankFile();

    bool deleteOnDisk(){
	if(next != 0){
	    next->deleteOnDisk();
	    delete next;
            next = 0;
	}
	close();
	int result = FileUtil::unlink(  this->mFileName.c_str()  );
	return (result == 0);	
    }

    bool insert(vector<int> key, DataType val);

    DataType get(const vector<int>& key) const;

    bool undefinedAt(const vector<int>& key) const;

    /*BucketData<DataType> get(int i) const;*/

    vector<int> getKey(int i) const;

    DataType getVal(int i) const;
    
    void setKey(int i, vector<int> & key);
    
    void setVal(int i, DataType val);
    
    BucketIterator<DataType>* iterator() const {return new BucketIterator<DataType>(this);}
    
    string getFileName(){return mFileName;}

    void printFileNames(){
	Bucket<DataType> * current = this;
	while(current != 0) {
	    cout << getFileName()<<endl;
	    current = current->next;
	}
	}
    
    void fillWithZeros(){
	if(next != 0){
	    next->deleteOnDisk();
	    delete next;
	    next = 0;
	}
	blankFile();
	close();
	open();
    }
 protected:
    int mNumberOfDimensions;
    int mBucketSize;
    int mIndex, mDepth;
/*    vector<BucketData<DataType> > data;*/
    int hash;
/*    int numItems;*/
    Bucket<DataType> *next;

    int mFileDescriptor; //file descriptor
    char *mData;// pointer to data
    
    string mFileName, mBaseFileName;

    int * mRefCount;

    enum {verbose = 0};

    inline uint64 getNumBytes(){
	return (sizeof(int) + (sizeof(int) * this->mNumberOfDimensions + sizeof(DataType)) * this->mBucketSize);
    }


/*    friend class BucketData<DataType>; */
    friend class BucketIterator<DataType>; 
    
};

template<class DataType>
void Bucket<DataType>::open() throw(std::ifstream::failure){
//  cout << "[Bucket] opening... "<<endl;
    assert (mData == 0);
    //copied from bigarray.h directly
    //do I completely understand it... not even close!
    
    FileUtil::mkdir_parentsonly(mFileName.c_str());
    if(!FileUtil::fileExists(mFileName.c_str()) || 
      (FileUtil::getFileSize(mFileName.c_str()) != this->getNumBytes())){
    // make sure the path the file exists
    //                
    const char * path = mFileName.c_str();
    char * shorter_path = strdup(path);
    char * p = strrchr(shorter_path, '/');// find the last occurance of '/'
    if(p == 0 || p == path) free(shorter_path);
    else {
      p[0] = 0;// set blank
      if(verbose) cout << "Creating directory "<<shorter_path << endl;
      int returnvalue = FileUtil::mkdir_parents(shorter_path);// should check return value
      if(verbose) cout << " Return value = " << returnvalue << endl;
      free(shorter_path);
    }
    // create the file
    blankFile();

  }
  FileUtil::mkdir_parentsonly(mFileName.c_str());
  mFileDescriptor = ::open(mFileName.c_str(), O_RDWR | O_CREAT);
  if(mFileDescriptor == -1 ) {
      stringstream strs;
      strs << "Failed to open file "<< mFileName<< " in R/W mode error code is " << errno<< " (see errno.h)";
      //string errormsg;
      //strs >>  errormsg ; 
      throw ifstream::failure(strs.str());
  }
  // I strongly suspect that if the volume is too big, the next line will fail
  // For now, just make the code fail until someone can check this out
  mData = (char *)mmap((caddr_t)0, this->getNumBytes(), 
			   PROT_READ | PROT_WRITE, MAP_SHARED, mFileDescriptor, 0);
//cout << "[Bucket] mData is mapped! "<<endl;
//cout <<" numBytes = "<<this->getNumBytes() <<endl;
  if ( mData == MAP_FAILED) {
      stringstream strs;
      strs << "Failed to map file "<< mFileName<< " error code is " << errno<< " (see errno.h)";
      //string errormsg;
      //strs >>  errormsg ; 
      throw ifstream::failure(strs.str());
  }

  if(this->getNumItems() > this->mBucketSize){
      next = new Bucket(this->mNumberOfDimensions, this->mBucketSize, this->hash, this->mIndex, this->mBaseFileName, this->mDepth+1);
      
  }

//  cout << "[Bucket.h] Finished opening." << endl;


}

template<class DataType>
void Bucket<DataType>::close(){
    if (mData != NULL)  {
//	cout << "UNMAPPING!!!!!!!!!!!!!!!"<<mFileName <<endl;
	if(next != 0){
	    next->close();
	    delete next;
	    next = 0;
	}


	munmap(mData, this->getNumBytes());
	mData = NULL;
	::close(mFileDescriptor);
    }
}

template <class DataType>
void Bucket<DataType>::blankFile() {
    //cout << " blanking the file " << endl;
    int * key = new int[this->mNumberOfDimensions];
    DataType * val = new DataType[1];
    memset(key,0,this->mNumberOfDimensions * sizeof(int));
    memset(val,0,sizeof(DataType));
    
    if(verbose) cout << "Blanking "<<mFileName << endl;
    ofstream FileStream( mFileName.c_str(),std::ios::binary |  std::ios::trunc);
    if(!FileStream.is_open()) {
	cout << " I could not create/write to " << mFileName << endl;
    }
    assert(FileStream.is_open());
    assert(!FileStream.fail());
    //FileStream.seekp(0);
    FileStream.write((char *) key, sizeof(int));
    for(int times = 0; times < this->mBucketSize; ++times){
	FileStream.write((char *) key,this->mNumberOfDimensions * sizeof(int));
	FileStream.write((char *) val,sizeof(DataType));	
    }
    assert(! FileStream.fail());
    FileStream.close();
    delete[] key;
    delete val;
}

template<class DataType>
vector<int> Bucket<DataType>::getKey(int i) const{
    vector<int> ret(this->mNumberOfDimensions);
    uint64 baseOffset = i * (sizeof(int) * this->mNumberOfDimensions + sizeof(DataType)) + sizeof(int);
    for(int k = 0; k < this->mNumberOfDimensions; k++){
	uint64 offset = baseOffset + k * sizeof(int);
	ret[k] = *((int *)(mData+offset));
    }			 
    return ret;
}
template<class DataType>
void Bucket<DataType>::setKey(int i, vector<int> & key){
    uint64 baseOffset = i * (sizeof(int) * this->mNumberOfDimensions + sizeof(DataType)) + sizeof(int);
    for(int k = 0; k < this->mNumberOfDimensions; k++){
	uint64 offset = baseOffset + k * sizeof(int);
	*((int *)(mData + offset)) = (key[k]);
    }			 
}

template<class DataType>
DataType Bucket<DataType>::getVal(int i) const{
    uint64 offset = (i+1) * (sizeof(int) * this->mNumberOfDimensions + sizeof(DataType)) + sizeof(int) - sizeof(mNumberOfDimensions);
    return *((DataType *)(mData+offset));
}
template<class DataType>
void Bucket<DataType>::setVal(int i, DataType val){
    uint64 offset = (i+1) * (sizeof(int) * this->mNumberOfDimensions + sizeof(DataType)) + sizeof(int) - sizeof(mNumberOfDimensions);
    *((DataType *)(mData + offset)) = val;
}

template<class DataType>
bool Bucket<DataType>::insert(vector<int> key, DataType val){
    for(int i = 0; i < min(this->getNumItems(),this->mBucketSize); i++){
	if(key == this->getKey(i)){
	    this->setVal(i,val);
	    return false;
	}
    }
    
    if(this->getNumItems() >= this->mBucketSize){
	
	if(this->getNumItems() == this->mBucketSize){
	    if(this->next != 0){ //sanity check
		next->deleteOnDisk();
		delete next;
		next = 0;
	    }
	    this->next = new Bucket(this->mNumberOfDimensions, this->mBucketSize, this->hash, this->mIndex, this->mBaseFileName, this->mDepth+1);
	    if(verbose) cout << "Created an overflow bucket. NumItems = " << this->getNumItems() << " BucketSize = " << this->mBucketSize << " Depth = " << mDepth << endl;
	}
	bool ret = next->insert(key,val);
	if(ret) *((int *)(mData)) = (int)(getNumItems() + 1);
	return ret;
    }else{
	int index = this->getNumItems();
	this->setKey(index,key);
	this->setVal(index,val);
//	cout << "Inserted " << this->data.size() << endl;
	*((int *)(mData)) = (int)(getNumItems() + 1);

	return true;
    }

}

template<class DataType>
DataType Bucket<DataType>::get(const vector<int>& key) const{
//    cout<< "size of data at beginning of get method " << data.size() << endl;
    for(int i = 0; i < min(this->getNumItems(),this->mBucketSize); i++){
//	cout << key[0] << ", " << key[1] << endl;
	if(key == this->getKey(i)){
	    return this->getVal(i);
	}
    }
    if(this->next != 0) return next->get(key);
    return 0;
}

template<class DataType>
bool Bucket<DataType>::undefinedAt(const vector<int>& key) const{
    for(int i = 0; i < min(this->getNumItems(),this->mBucketSize); i++){
	if(key == this->getKey(i)){
	    return false;
	}
    }
    if(this->next != 0) return next->undefinedAt(key);
    return true;
}

/*template<class DataType>
BucketData<DataType> Bucket<DataType>::get(int i) const{
    if(i < this->mBucketSize){
	return this->data[i];
    }
    else{
	return next->get(i-mBucketSize);
    }
}
*/

/*
template<class DataType>
class BucketData{
 public:
    BucketData():key(0),value(0){;}

    BucketData(vector<int>& x, DataType y): key(x), value(y){;}

    inline vector<int> getKey() const {return key;}
    inline int getKey(int i) const {return key[i];}
    inline DataType getValue() const {return value;}
    
    bool operator == (BucketData b2){	
	vector<int> key2 = b2.getKey();
	if(this->key.size() != key2.size()) return false;
	for(uint i = 0; i < this->key.size(); i++){
	    if(this->key[i] != key2[i]) return false;
	}
	return true;
    }
 private:
    vector<int> key;
    DataType value;
};

*/

template<class DataType>
class BucketIterator{
 public:
    BucketIterator(const Bucket<DataType> *b):data(b){
	this->count = -1;
	
};
    
    inline DataType val() const {return data->getVal(count);}
    inline vector<int> key() {return data->getKey(count);}
    inline int key(int i) {
	vector<int> k = data->getKey(count);
	return k[i];
    }

    inline bool next(){
	count++;
//	cout << count << " " << data->getNumItems() << endl;
	if(count >= data->getNumItems())
	    return false;
	if(count == data->getSize()){
	    data = data->getNext(); 
	    count = 0;
	}
	return true;	
}

 protected:

 private:
    const Bucket<DataType> *data;
    int count;
};

#endif
