//Author : Steven Keith can be blamed for this...

#ifndef LHWRBI_H
#define LHWRBI_H

#include <string>
#include <fstream>
#include <cassert>
#include <map>
#include <list>
#include <set>
#include <sstream>
#include <iostream>
#include <stdlib.h>
#include <limits.h>
#include <stdio.h>
#include "fileutil.h"
#include "multidimensionalarray.h"
#include "hint.h"
#include "bucket.h"

using namespace std;

template<class DataType>
class LinearHashCubeIterator; 
template<class DataType>
class RangeBucketIterator;

template<class DataType>
class LinearHashCube:public MultidimensionalArray<DataType> {
    
 public:
    LinearHashCube( int NumberOfDimensions, string filename = "LinearHashCubeTemp.bin", Hint h = Hint()) throw(std::ifstream::failure);
    LinearHashCube( int NumberOfDimensions, string filename, int bucketSize, int numberHashingBits = 15) throw(std::ifstream::failure);
    
    virtual ~LinearHashCube(){
        storage.clear();
        this->close();
        delete mData;
    }

    int maxInBucket(){
//        cerr << endl << endl;
        int ret = 0;
        for(int i = 0; i < (int)storage.size(); i++){
            if(storage[i].getNumItems() > ret) ret = storage[i].getNumItems();
//            cerr << storage[i].getNumItems() << " ";
        }
//        cerr << endl;
        return ret;
    }
    
    string getName() const {return this->mFilename;}
    
    int getNumberOfDimensions() const {return this->mNumberOfDimensions;}
    
    int getNumberOfBuckets() const {return this->getNumBuckets();}
    
    int getNumHashingBits() const {return this->numHashingBits;}

    void open() throw(std::ifstream::failure);
    
    inline uint64 getNumBytes() const{return 6 * sizeof(int);}
    
    void tempPrintGraph(){
        for(int i = 0; i < this->getNumberOfBuckets(); i++){
            cout << i << " " << this->storage[i].getNumItems() << endl;
        }
    }

    list<string>* getFileNames(){
	list<string>* ret = new list<string>();
	ret->push_back(this->getName());
	for(int i = 0; i < this->getNumBuckets(); i++){
	    storage[i].getFileNames(ret);
	}
	return ret;
    }

    void close(){
	if (mData != NULL)  {
	    for(int i = 0; i < this->getNumBuckets(); i++){
                if(verbose) storage[i].printFileNames();
		storage[i].close();
	    }
	    munmap(mData, this->getNumBytes());
	    mData = 0;
	    ::close(mFileDescriptor);
   	    if(verbose) cout << "Clearing storage..." << endl;
/*	    if(verbose){
		for(int i = 0; i < this->getNumBuckets(); i++){
		    storage[i].printFileNames();
		}
		}*/
	    storage.clear();
	}
    }    

    void blankFile();

    //does nothing
    void sync(){;}
    
    //deletes all values in the cube
    void fillWithZeros();

    virtual bool deleteOnDisk() {
        if(verbose)cout<<"linearhash deleteOnDisk"<<endl;
        if(verbose)cout <<"linearhash about to delete buckets regarding file "<< mFilename << endl;
        if(verbose)cout<<" number of buckets = " << this->getNumBuckets() << endl;
        const int num = this->getNumBuckets();
	bool result = true;
	for(int i = 0; i < num; i++){
	    if (verbose) cout << "Deleting Bucket " << i << endl;
	    result = result && storage[i].deleteOnDisk();
	    if(verbose)cout<< "i = "<< i<<" number of buckets = " << this->getNumBuckets() << endl;
	}
	//this->close();
        this->close();
        if(verbose)cout <<"linearhash about to delete file "<< mFilename << endl;
	int res = FileUtil::unlink(  mFilename.c_str()  );
	return (res == 0 && result);
    }    

    //inserts a value into the cube
    inline void put(const vector<int>& pos, DataType value);
    inline void put(const int pos[], DataType value);
    
    inline DataType get( const vector<int>& pos) const;
    inline DataType get( const int pos[]) const;
    
    inline bool undefinedAt( const vector<int>& pos) const;
    inline bool undefinedAt( const int pos[]) const;
    
    BucketIterator<DataType> getIter(int x) const {return *storage[x].iterator();}
    Bucket<DataType> getBucket(int i) const{return storage[i];}

    typename MultidimensionalArray<DataType>::
	IteratorPtr iterator(const vector<int>& start, const vector<int>& end) const {
        return typename MultidimensionalArray<DataType>::IteratorPtr(new LinearHashCubeIterator<DataType>(*this,start,end));
    }

    typename MultidimensionalArray<DataType>::
	IteratorPtr iterator() const {
        return typename MultidimensionalArray<DataType>::IteratorPtr(new LinearHashCubeIterator<DataType>(*this));
    }

    typename MultidimensionalArray<DataType>:: IteratorPtr sliceIterator(int dim, int slice_at) {
	vector<int> s(this->mNumberOfDimensions),e(this->mNumberOfDimensions);
	for(uint i = 0; i < s.size(); i++){
	    s[i] = 0;
	}
	for(uint i = 0; i < e.size(); i++){
	    e[i] = numeric_limits<int>::max();
	}
	s[dim] = slice_at;
	e[dim] = slice_at + 1;
	return this->iterator(s,e);
    }
    
    typename MultidimensionalArray<DataType>:: IteratorPtr diceIterator(int dim, int dice_at) const {
	vector<int> s(this->mNumberOfDimensions),e(this->mNumberOfDimensions);
	for(uint i = 0; i < s.size(); i++){
	    s[i] = 0;
	}
	for(uint i = 0; i < e.size(); i++){
	    e[i] = numeric_limits<int>::max();//2 << (30/this->mNumberOfDimensions >> 1) - 1;
	}
	e[dim] = dice_at;
	return this->iterator(s,e);
    }



    //the following functions should likely be private... will move them
    //when testing is complete.

    int interleave(const vector<int> key) const;
    int interleave(const int key[]) const;
	
    inline float measureUsage(){
	return (float)this->getNumRecords()/(float)(this->storage.size()*this->getRecordsPerBucket());
    }
    inline int getNumberOfRecords() const{return this->getNumRecords();}
 protected:
    
    inline int getNumRecords() const {return *((int *)(mData));}
    inline int getRecordsPerBucket() const {return *((int *)(mData+sizeof(int)));}
    inline int getS() const{return *((int *)(mData+2*sizeof(int)));}
    inline int getNumBuckets() const{return *((int *)(mData+3*sizeof(int)));}
    inline float getTau() const{return (float)(* ((int *)(mData+4*sizeof(int))))/(float)( *((int *)(mData+5*sizeof(int))));}
    inline int getNumberOfHashingBits() const{return *((int *)(mData+6*sizeof(int)));}

    inline void setNumRecords(int val) {
	uint64 offset = 0;
	*((int *)(mData + offset)) = (int)(val);
    }
    inline void setRecordsPerBucket(int val) {
	uint64 offset = sizeof(int);
	*((int *)(mData + offset)) = (int)(val);
    }
    inline void setS(int val) {
	uint64 offset = 2*sizeof(int);
	*((int *)(mData + offset)) = (int)(val);
    }
    inline void setNumBuckets(int val) {
	uint64 offset = 3*sizeof(int);
	*((int *)(mData + offset)) = (int)(val);
    }
    inline void setTau(int num, int den) {
	uint64 offset = 4*sizeof(int);
	*((int *)(mData + offset)) = (int)(num);
	offset = 5*sizeof(int);
	*((int *)(mData + offset)) = (int)(den);
    }
    inline void setNumHashingBits(int val) {
	uint64 offset = 6*sizeof(int);
	*((int *)(mData + offset)) = (int)(val);
    }
    


 private:

//    uint getNumBytes(){return (sizeof(int) * 3 + sizeof(float));}

    void split();
    
    virtual MultidimensionalArray<DataType>* createNew(vector<int> shape,string file, Hint hint = Hint()) const  throw(std::ifstream::failure){
	return new LinearHashCube(shape.size(),file,hint);
	}

    int mNumberOfDimensions;
    Hint mHint;
    const string mFilename;
    
    int mBucketSize;

    string PREFIX;

    vector<Bucket<DataType> > storage;

    int mFileDescriptor; //file descriptor
    char *mData;// pointer to data
    int numHashingBits;
    enum {verbose=0};

    friend class LinearHashCubeIterator<DataType>; 
    friend class RangeBucketIterator<DataType>;

};


template<class DataType>
LinearHashCube<DataType>::LinearHashCube( int NumberOfDimensions, string filename, Hint h) throw(std::ifstream::failure):MultidimensionalArray<DataType>(vector<int>(NumberOfDimensions,this->INVALID_SIZE)), mNumberOfDimensions(NumberOfDimensions), mHint(h),mFilename(filename), mBucketSize(100),PREFIX("[lhwrbi"),storage(vector<Bucket<DataType> >(0)), mFileDescriptor(0), mData(0), numHashingBits(15){
    this->mBucketSize=h.getBucketSize();
    this->numHashingBits=h.getNumHashingBits();
    this->open();
}

template<class DataType>
LinearHashCube<DataType>::LinearHashCube( int NumberOfDimensions, string filename, int bucketSize, int numberHashingBits) throw(std::ifstream::failure):MultidimensionalArray<DataType>(vector<int>(NumberOfDimensions,this->INVALID_SIZE)), mNumberOfDimensions(NumberOfDimensions), mHint(),mFilename(filename), mBucketSize(bucketSize),PREFIX("[lhwrbi"),storage(vector<Bucket<DataType> >(0)), mFileDescriptor(0), mData(0), numHashingBits(numberHashingBits){
    this->open();
	
}

template<class DataType>
void LinearHashCube<DataType>::open() throw(std::ifstream::failure){
    if(verbose) cout << "[lhwrbi] opening... "<<endl;
    assert (mData == 0);
    //copied from bigarray.h directly
    //do I completely understand it... not even close!
    bool newFile = false;    
    
    FileUtil::mkdir_parentsonly(this->mFilename.c_str());
    if(!FileUtil::fileExists(this->mFilename.c_str()) || 
       (FileUtil::getFileSize(this->mFilename.c_str()) != this->getNumBytes())){
	if(verbose){
	    if(!FileUtil::fileExists(this->mFilename.c_str())){
		cout << "the file does not exist!" << endl;
	    }
	    else{
		cout << FileUtil::getFileSize(this->mFilename.c_str()) << endl;
		cout << this->getNumBytes() << endl;
	    }
	}
	newFile = true;
	// make sure the path the file exists
	//                
	const char * path = mFilename.c_str();
	char * shorter_path = strdup(path);
	char * p = strrchr(shorter_path, '/');// find the last occurance of '/'
	if(p == 0 || p == path) free(shorter_path);
	else {
	    p[0] = 0;// set blank
	    if(verbose) cout << "Creating directory "<<shorter_path << endl;
	    int returnvalue = FileUtil::mkdir_parents(shorter_path);// should check return value
	    if(verbose) cout << " Return value = " << returnvalue << endl;
	    free(shorter_path);
	}
	// create the file
	this->blankFile();
	
    }
    FileUtil::mkdir_parentsonly(mFilename.c_str());
    mFileDescriptor = ::open(mFilename.c_str(), O_RDWR | O_CREAT);
    if(mFileDescriptor == -1 ) {
	stringstream strs;
	strs << "Failed to open file "<< mFilename<< " in R/W mode error code is " << errno<< " (see errno.h)";
	
	throw ifstream::failure(strs.str());
    }
    // I strongly suspect that if the volume is too big, the next line will fail
    // For now, just make the code fail until someone can check this out
    mData = (char *)mmap((caddr_t)0, this->getNumBytes(), 
			 PROT_READ | PROT_WRITE, MAP_SHARED, mFileDescriptor, 0);
    if(verbose) cout << "[lhwrbi.h] mData is mapped! "<<endl;
    if(verbose) cout <<" numBytes = "<<this->getNumBytes() <<endl;
    if ( mData == MAP_FAILED) {
	stringstream strs;
	strs << "Failed to map file "<< mFilename<< " error code is " << errno<< " (see errno.h)";
	throw ifstream::failure(strs.str());
    }
    if(newFile){
	this->setNumRecords(0);
	this->setRecordsPerBucket(this->mBucketSize);
	this->setS(0);
	this->setNumBuckets(1);
	this->setTau(7,10);
	this->setNumHashingBits(this->numHashingBits);
	this->storage.push_back(Bucket<DataType>(this->mNumberOfDimensions, this->getRecordsPerBucket(), 0,0,this->mFilename));
	if(verbose) cout << "No buckets found... added a single empty bucket" << endl;	
    }
    else{
	int numBuckets = this->getNumBuckets();
	if (numHashingBits != this->getNumberOfHashingBits()){
	  stringstream strs;
	  strs << "Number of Hashing Bits does not match!";
	  throw ifstream::failure(strs.str());
	}
	if(verbose) cout << "numBuckets = " << numBuckets << endl;
	// OFK is worried about the rounding aspects of this. 
        // but, since it's just a hash, being one too small might be fine
	// Steve is assuming this will always round down as two different
	// hashes are almost always in play... one the floor of this fraction
	// and one the ceil.
	int hash = (int)(log10((double)numBuckets)/log10(2.0));
	if(verbose) cout << "hash = " << hash << endl;
	int threshold = 2 << hash >> 1;
	int outer = numBuckets - threshold;
	int inner = threshold - outer;
	int count = 0;
	int i = 0;
	while(count < outer){
	    storage.push_back(Bucket<DataType>(this->mNumberOfDimensions, this->getRecordsPerBucket(), hash+1,i,this->mFilename));	    
	    count ++;
	    i ++;
	}
	count = 0;
	while(count < inner){
	    storage.push_back(Bucket<DataType>(this->mNumberOfDimensions, this->getRecordsPerBucket(), hash,i,this->mFilename));	    
	    count ++;
	    i ++;
	}
	count = 0;
	while(count < outer){
	    storage.push_back(Bucket<DataType>(this->mNumberOfDimensions, this->getRecordsPerBucket(), hash+1,i,this->mFilename));	    
	    count ++;
	    i ++;
	}

/*	if(verbose && this->getNumBuckets() > 0) cout << "Buckets found... opening them" << endl;
	for(int i = 0; i < this->getNumBuckets(); i++){
	    if(i > threshold && i < numBuckets - threshold - 1 ){
		storage.push_back(Bucket<int>(this->mNumberOfDimensions, this->getRecordsPerBucket(), hash,i,this->mFilename));
	    }
	    else{
		storage.push_back(Bucket<int>(this->mNumberOfDimensions, this->getRecordsPerBucket(), hash+1,i,this->mFilename));
	    }
	    }
*/
    }

    if(verbose) cout << "[lhwrbi.h] Finished opening." << endl;
}

template <class DataType>
void LinearHashCube<DataType>::blankFile() {
    int * key = new int[7];
    memset(key,0,7 * sizeof(int));
    
    if(verbose) cout << "Blanking "<<mFilename << endl;
    ofstream FileStream( mFilename.c_str(),std::ios::binary |  std::ios::trunc);
    if(!FileStream.good()) {
	cout << " I could not create/write to " << mFilename << endl;
    }
    assert(FileStream.good());
    assert(!FileStream.fail());

    FileStream.write((char *) key,7 * sizeof(int));

    assert(! FileStream.fail());
    FileStream.close();
    delete[] key;
    
}



template<class DataType>
void LinearHashCube<DataType>::fillWithZeros(){
    this->deleteOnDisk();
    this->open();
}

template<class DataType>
int LinearHashCube<DataType>::interleave(const vector<int> key) const{
    vector<int> key2(mNumberOfDimensions);
    for(int i = 0; i < mNumberOfDimensions; i++){
	key2[i] = key[i];
    }
	
    int val = 0;
    int precision = (int) ceil(this->numHashingBits / (float)this->mNumberOfDimensions);
//    int max = (2 << precision >> 1);
//    for(int i = 0; i < mNumberOfDimensions; i++){
//	key2[i] = key2[i] % max;
//    }
//    int count = 29;
    int count = this->numHashingBits - 1;

    for(int i = 0; i < precision; i++){
	for(int j = this->mNumberOfDimensions - 1; j >= 0; j--){
	    if((precision - i) * this->mNumberOfDimensions - (this->mNumberOfDimensions - j) > count) continue;
	    val += ((key2[j] % 2)* (2 << count >> 1));
	    key2[j] = key2[j] / 2;
	    count--;
	}
    }
    return val;
}

template<class DataType>
int LinearHashCube<DataType>::interleave(const int key[]) const{
    vector<int> key2(mNumberOfDimensions);
    for(int i = 0; i < mNumberOfDimensions; i++){
	key2[i] = key[i];
    }
	
    int val = 0;
    int precision = (int) ceil(this->numHashingBits / (float)this->mNumberOfDimensions);
//    int max = 2 << precision >> 1;
//    for(int i = 0; i < mNumberOfDimensions; i++){
//	key2[i] = key2[i] % max;
//    }
    int count = this->numHashingBits - 1;

    for(int i = 0; i < precision; i++){
	for(int j = this->mNumberOfDimensions - 1; j >= 0; j--){
	    if((precision - i) * this->mNumberOfDimensions - (this->mNumberOfDimensions - j) > count) continue;
	    val += ((key2[j] % 2)*(2 << count >> 1));
	    key2[j] = key2[j] / 2;
	    count--;
	}
    }
    return val;
}

template<class DataType>
void LinearHashCube<DataType>::put(const vector<int>& pos, DataType value){
    int val = interleave(pos);
    int index = val % (2 << storage[0].getHash() >> 1);
    int numBuckets = this->storage.size();
    if(index >= numBuckets){
	index = val % (2 << storage[0].getHash() >> 2);
    }
//    cout << "Index = " << index << endl;

    if(verbose)cout << "Trying a put in bucket " << index << endl;
    if(storage[index].insert(pos,value)){
//	cout << "Inserting into bucket " << index << endl;
	this->setNumRecords(this->getNumRecords() + 1);
    }
    if(verbose)cout << "Inserted into " << index << " NumRecords = " << this->getNumRecords() << " Usage = " << this->measureUsage() << " Tau = " << this->getTau() << endl;
    if(measureUsage() > this->getTau()){
	if(verbose)cout << "SPLIT!" << endl;
	split();
    }
}

template<class DataType>
void LinearHashCube<DataType>::put(const int pos[], DataType value){
    vector<int> key(this->mNumberOfDimensions);
    for(int i = 0; i<this->mNumberOfDimensions; i++){
	key[i] = pos[i];
    }
    return this->put(key,value);
}
   
template<class DataType>
DataType LinearHashCube<DataType>::get(const vector<int>& pos) const{
    int val = interleave(pos);
    int index = val % (2 << storage[0].getHash() >> 1);
    int numBuckets = this->storage.size();
    if(index >= numBuckets){
	index = val % (2 << storage[0].getHash() >> 2);
    }
    return storage[index].get(pos);
}

template<class DataType>
DataType LinearHashCube<DataType>::get(const int pos[]) const{
    vector<int> key(this->mNumberOfDimensions);
    for(int i = 0; i<this->mNumberOfDimensions; i++){
	key[i] = pos[i];
    }
    return this->get(key);
}

template<class DataType>
bool LinearHashCube<DataType>::undefinedAt(const vector<int>& pos) const{
    int val = interleave(pos);
    int index = val % (2 << storage[0].getHash() >> 1);
    int numBuckets = this->storage.size();
    if(index >= numBuckets){
	index = val % (2 << storage[0].getHash() >> 2);
    }
    return storage[index].undefinedAt(pos);
}

template<class DataType>
bool LinearHashCube<DataType>::undefinedAt(const int pos[]) const{
    vector<int> key(this->mNumberOfDimensions);
    for(int i = 0; i<this->mNumberOfDimensions; i++){
	key[i] = pos[i];
    }
    return this->undefinedAt(key);
}

template<class DataType>
void LinearHashCube<DataType>::split(){
    uint index1 = this->getS();
    uint index2 = this->storage.size();

    int hash = this->storage[index1].getHash() + 1;

    vector<vector<int> > keys;
    vector<DataType> vals;
    
    BucketIterator<DataType> *splitIter = storage[index1].iterator();
    while(splitIter->next()){
	keys.push_back(splitIter->key());
	vals.push_back(splitIter->val());
    }
    
    delete splitIter;
    
    this->storage[index1].fillWithZeros();
    this->storage[index1].setHash(hash);
    
//    storage[index1] = (Bucket<DataType>(this->mNumberOfDimensions, this->getRecordsPerBucket(), hash, (int)index1, this->mFilename));
    this->storage.push_back(Bucket<DataType>(this->mNumberOfDimensions, this->getRecordsPerBucket(), hash, (int)index2, this->mFilename));
    
    if(verbose)cout << "Checking to insure both buckets are now empty " << storage[index1].getNumItems() << " " << this->storage[index2].getNumItems() << endl;
    if(verbose)cout << "Checking to insure all buckets are still open and defined..."<<endl;
    if(verbose){
	for(uint i = 0; i < this->storage.size(); i++){
	    cout << i << " " << this->storage[i].getNumItems() << endl;
	}
    }
    if(verbose)cout << "Checking for unique file names " << storage[index1].getFileName() << " " << this->storage[index2].getFileName() << endl;
 
    if(verbose)cout << "Splitting " << keys.size() << " values into "<< index1 << " and " << index2 <<endl;
    //splitting the contents between the two buckets
    for(uint i = 0; i < keys.size(); i++){

	if(interleave(keys[i]) % (2 << hash >> 1) == (int)index1){
	    this->storage[index1].insert(keys[i], vals[i]);
//	    cout << "one" << endl;
	}
	else{
	    this->storage[index2].insert(keys[i], vals[i]);
//	    cout << "two" << endl;
	}
    }//end for

    if(verbose)cout << "Bucket " << index1 << " has " << this->storage[index1].getNumItems() << " items" << endl;
    if(verbose)cout << "Bucket " << index2 << " has " << this->storage[index2].getNumItems() << " items" << endl;

    //incremement s or reset it to 0
    this->setS(index1 + 1);
    if(this->getS() == (2 << hash >> 2)){
	this->setS(0);
    }
    //increment number of buckets
    this->setNumBuckets(this->getNumBuckets()+1);

    if(verbose){
	for(uint i = 0; i < this->storage.size(); i++){
	    cout << storage[i].getHash() << " " ;
	}
	cout << endl;
    }

    return;
}


template <class DataType>
class LinearHashCubeIterator:public MultidimensionalIterator<DataType> {

 public:
    LinearHashCubeIterator(const LinearHashCube<DataType> & owner,const vector<int>& start, const vector<int>& end) :
	MultidimensionalIterator<DataType>(),
	mStart(start), 
	mEnd(end),
	mOwner(owner),  
	mKey(mStart), 
	mCurrentValue(0),
	iter(0),
	mSearchTheseBuckets(0),
        missed(0)
	{
//	    cout << "Max Hash = " << mOwner.getBucket(0).getHash() << endl;
	    if(verbose){
		cout << "start ";
		for(uint i = 0; i < start.size(); i++){
		    cout << start[i] << " ";
		}
		cout << endl;
		cout << "end ";
		for(uint i = 0; i < end.size(); i++){
		    cout << end[i] << " ";
		}
		cout << endl;
	    }
	    vector<int> rbiStart(this->mEnd.size());
	    vector<int> rbiEnd(this->mEnd.size());
	    vector<int> interleaveMax(this->mEnd.size());
	    for(int i = 0; i < this->mOwner.getNumHashingBits(); i++){
	      ++ interleaveMax[i % this->mEnd.size()];
	    }
	    for(int i = 0; i < (int)this->mEnd.size(); i++){
	      interleaveMax[i] = (2 << interleaveMax[i] >> 1);
	      if(verbose) cout << "InterleaveMax[" << i << "] = " << interleaveMax[i] << endl;
	    }

	    for(uint i = 0; i < this->mEnd.size(); i++){
		if(this->mEnd[i] - this->mStart[i] == 1){
		    rbiStart[i] = this->mStart[i];
		}
		else if(this->mStart[i] < this->mEnd[i]  
			&& this->mEnd[i] - this->mStart[i] < interleaveMax[i]
			&& this->mStart[i] % interleaveMax[i] < this->mEnd[i] % interleaveMax[i]){
		  rbiStart[i] = this->mStart[i] % interleaveMax[i];
		  if(verbose)cout << "rbiStart[" << i << "] = " << rbiStart[i] << endl;
		}
		else{
		    rbiStart[i] = 0;
		}
	    }


	    for(uint i = 0; i < this->mEnd.size(); i++){
		if(this->mEnd[i] - this->mStart[i] == 1){
		    rbiEnd[i] = this->mStart[i];
		}		
		else if(this->mStart[i] < this->mEnd[i]  
			&& this->mEnd[i] - this->mStart[i] < interleaveMax[i]
			&& this->mStart[i] % interleaveMax[i] < this->mEnd[i] % interleaveMax[i]){
		  rbiEnd[i] = this->mEnd[i] % interleaveMax[i];
		  if(verbose)cout << "rbiEnd[" << i << "] = " << rbiEnd[i] << endl;
		}
		else{
		  rbiEnd[i] = interleaveMax[i] - 1;
		}
	    }
	    
	    RangeBucketIterator<DataType> rbi(this->mOwner, this->mStart, rbiEnd);
	    list<int> temp = rbi.getBucketList();
	    mSearchTheseBuckets.splice(mSearchTheseBuckets.end(),temp);
	    mSearchTheseBuckets.push_back(-1);
	    list<int>::const_iterator it(mSearchTheseBuckets.begin());
	    
	    if(verbose) {
		cout<< "Searching these buckets" << endl;
		while(it != mSearchTheseBuckets.end()){
		    cout << "bucket #" << *it << endl;
		    it ++;
		}
	    }
	    this->currentBucket = *mSearchTheseBuckets.begin();
	    this->mSearchTheseBuckets.pop_front();
	    if(currentBucket >= 0){
		this->iter= mOwner.getIter(currentBucket);
	    }
	}

    LinearHashCubeIterator(const LinearHashCube<DataType> & other):
	MultidimensionalIterator<DataType>(),
	mStart(other.getShape().size()),
	mEnd(other.getShape().size()), 
	mOwner(other), 
	mKey(other.getShape().size()),
	mCurrentValue(0),
	iter(0),
	mSearchTheseBuckets(0),
        missed(0)
	{
	    for(uint i = 0; i < mStart.size(); i++){
		this->mStart[i] = 0;
	    }
	    for(uint i = 0; i < mEnd.size(); i++){
		this->mEnd[i] = numeric_limits<int>::max();
	    }

	    for(int i = 0; i < this->mOwner.getNumberOfBuckets(); i++){
		this->mSearchTheseBuckets.push_back(i);
	    }
	    this->mSearchTheseBuckets.push_back(-1);
	    this->currentBucket = *mSearchTheseBuckets.begin();
	    this->mSearchTheseBuckets.pop_front();
	    if(currentBucket >= 0){
		this->iter= mOwner.getIter(currentBucket);
	    }

	}
    
    LinearHashCubeIterator(const LinearHashCubeIterator<DataType> & other):
	MultidimensionalIterator<DataType>(),
	mStart(other.mStart), 
	mEnd(other.mEnd), 
	mOwner(other.mOwner), 
	mKey(other.mKey), 
	mCurrentValue(other.mCurrentValue),
	iter(other.iter),
	mSearchTheseBuckets(0),
        missed(other.missed)
	{
	    this->currentBucket = 0;
	    list<int> temp = other.mSearchTheseBuckets;
	    mSearchTheseBuckets.splice(mSearchTheseBuckets.end(),temp);
	}
    
    
    inline bool next(){
	if(mSearchTheseBuckets.size() == 0)
	    return false;
	
	bool alright = false;
	if(iter.next()){
	    while(!alright){
		alright = true;
		for(int i = 0; i < this->mOwner.getNumberOfDimensions(); i++){
		    int temp = iter.key(i);
		    if(!(mStart[i] <= temp && temp < mEnd[i])){
			alright = false;
                        missed++;
//                        cerr << missed << " ";
		    }
		}
		if(alright){
		    mCurrentValue = iter.val();
		    for(int i = 0; i < this->mOwner.getNumberOfDimensions(); i++){
			mKey[i] = iter.key(i);
		    }
		
		}
		else if(!iter.next()){
		    return this->next();
		}
	    }
	}
	else{
	    this->currentBucket = *mSearchTheseBuckets.begin();
	    this->mSearchTheseBuckets.pop_front();
	    while( this->mSearchTheseBuckets.size() > 0 && this->mOwner.storage[currentBucket].getNumItems() == 0){
		this->currentBucket = *mSearchTheseBuckets.begin();
		this->mSearchTheseBuckets.pop_front();
	    }
	    if(mSearchTheseBuckets.size() == 0) {
		return false;
	    }else{
		this->iter= mOwner.getIter(currentBucket); 
		return this->next();
	    }
	    
	}
	return true;
    }
    
    inline DataType value() const{
	return this->mCurrentValue;
    }
    
    inline vector<int>& key(){
	return mKey;
    }
    
    LinearHashCubeIterator<DataType> * clone() const {return new LinearHashCubeIterator<DataType>(*this);}
    
 protected:
    int currentBucket;
    vector<int> mStart, mEnd;
    const LinearHashCube<DataType> & mOwner;
    vector<int> mKey;
    DataType mCurrentValue;
    enum {verbose=0};
    BucketIterator<DataType> iter;
    list<int> mSearchTheseBuckets;
    int missed;
    
};

template<class DataType>
class RangeBucketIterator{

 public:
    
    RangeBucketIterator(const LinearHashCube<DataType> & owner,const vector<int>& start, const vector<int>& end): 
	mStart(start), 
	mEnd(end),
	mBuckets(0),
	mOwner(owner)
	{
	    this->query(0,1,mStart,mEnd);
	}
    
    
    void query(int index, int hash, vector<int> ll, vector<int> ur);

    list<int> getBucketList() const{return mBuckets;}
    
 protected:
    vector<int> mStart, mEnd;
    list<int> mBuckets;
    const LinearHashCube<DataType> & mOwner;

};

template<class DataType>
void RangeBucketIterator<DataType>::query(int index, int hash, vector<int> in_ll, vector<int> in_ur){
    if(index >= this->mOwner.getNumBuckets()){
	return;
    }
    list<int> returnMe(0);
    
    if(this->mOwner.getBucket(index).getHash() < hash){
	this->mBuckets.push_back(index);
	return;
    }
    int ll = this->mOwner.interleave(in_ll);
    int ur = this->mOwner.interleave(in_ur);
    int h1 = (ll % (2 << hash >>1) - ll % (2 << hash >> 2))/(2 << hash >> 2);
    int h2 = (ur % (2 << hash >>1) - ur % (2 << hash >> 2))/(2 << hash >> 2);
    int index1 = ll % (2 << hash >> 1);
    int index2 = ur % (2 << hash >> 1);    
   
    int numDims = this->mOwner.getNumberOfDimensions();
    int coordinate = (hash-1) % numDims;
    int blockSize = (2 << ((this->mOwner.getNumHashingBits()-hash)/numDims) >> 1);
    if(verbose){
	cout << hash << " " << this->mOwner.getBucket(index).getHash() << " " << index << " " << index1 << " " << index2 << " " << coordinate << " " << blockSize << " " << h1 << " " <<h2 <<endl;
	cout << ll << " ";
	for(uint i = 0 ; i < in_ll.size(); i++){
	    cout << in_ll[i] << " ";
	}
	cout << endl;
	cout << ur << " ";
	for(uint i = 0 ; i < in_ur.size(); i++){
	    cout << in_ur[i] << " ";
	}
	cout << endl;
    }
     //Does the lower left and upper right corners of the query rectangle
    //hash to the same bucket?
    if(h1 == h2){
	this->query(index + h1 * (2 << hash >> 2),hash+1,in_ll,in_ur);
	return;

    }//otherwise redefine each of the queries to fit in a single bucket
    else{
	vector<int> ur2(numDims), ll2(numDims);
	for(int i = 0; i < numDims; i++){
	    ur2[i] = in_ur[i];
	    ll2[i] = in_ll[i];
	}

//	int coordinate = (hash) % numDims;
	ur2[coordinate] = (in_ll[coordinate]/ blockSize + 1) * blockSize-1;
	ll2[coordinate] = ((in_ur[coordinate]-1)/blockSize ) * blockSize;
	this->query(index,hash+1,in_ll,ur2);
	this->query(index+(int)(2 << hash >> 2),hash+1,ll2,in_ur);
	return;
    }
}


#endif
