#include "lhwrbi.h"
#include "hint.h"
#include "bigarray.h"
#include "internalbigarray.h"
#include "btreecube.h"
#include "btreecube_rotated.h"
#include "cdbcube.h"
#include <time.h>
#include <cmath>
#include <stdlib.h>

void hardertest(){
    srand(0);
    int numDims = 5;
    int numBits = 30;
    int numInserts = 10000;
    vector<int> pos(numDims);
    clock_t start,finish;
    double seconds;
    int counter = 0;
    
    LinearHashCube<int> cube(numDims, "linearhashtest.bin", 100);
    cout << "Cube created.  Inserting " << numInserts << " values " << endl;
    start = clock();
    while(cube.getNumberOfRecords() < numInserts){
	counter++;
	if(cube.getNumberOfRecords() % 10000 == 0) cout << (((float)cube.getNumberOfRecords()/(float)numInserts) * 100) << "% complete" << endl;
	for(int j = 0; j < numDims; j++){
	    do{
	    pos[j] =(int)floor((double)rand()/RAND_MAX*(2<<(int)(numBits/numDims * 1.5)>>1)) - 1;
	    }while(pos[j] < 0);
//	    cout << pos[j] << endl;
	}
	cube.put(pos,1);
    }
    finish = clock();
    seconds =  (double)(finish - start) / CLOCKS_PER_SEC;    
    cout << "done in " << seconds << " seconds." << endl;
    cout << "Average " << (seconds / counter) << " seconds/put" << endl << endl;

    cout << "getting numBuckets... " << cube.getNumberOfBuckets() << endl;
    cout << "getting numNumberOfRecords..." << cube.getNumberOfRecords() << endl;
//    while(true);

    cout << "getting an iterator of everything and summing the entries..." << endl;;
    start = clock();    
    MultidimensionalArray<int>::IteratorPtr iter(cube.iterator());
    int count = 0;
    while(iter->next()){
	count++;
    }
    finish = clock();
    seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
    cout << "iterated through " << count << " items." << endl;
    cout << "done in " << seconds << " seconds." << endl << endl;

    cout << "closing and re-opening the file to test persistance..."<< endl;
    cube.close();
    start = clock();    
    cube.open();
    finish = clock();
    seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
    cout << "done in " << seconds << " seconds." << endl << endl;

    cout << "checking to make sure things load from files correctly..." << endl;
    cout << "getting numBuckets... " << cube.getNumberOfBuckets() << endl;
    cout << "getting numNumberOfRecords..." << cube.getNumberOfRecords() << endl;
    cout << "getting an iterator of everything and summing the entries..." <<endl;
    iter = cube.iterator();
    count = 0;
    while(iter->next()){
	count++;
    }
    cout << "iterated through " << count << " items." << endl << endl;

    cout << "re-seeding the random number generator... this should let us generate the same keys..." << endl << endl;
    srand(0);
    cout << "testing the get operation..."<<endl;
    count = 0;
    start = clock();
    while(count < counter){
	count++;
	if(count % 10000 == 0) cout << (((float)count/(float)numInserts) * 100) << "% complete" << endl;
	for(int j = 0; j < numDims; j++){
	    pos[j] =(int)floor((double)rand()/RAND_MAX*(2<<(numBits/numDims)>>1));
//	    cout << pos[j] << endl;
	}
	int temp = cube.get(pos);
    }
    finish = clock();
    seconds =  (double)(finish - start) / CLOCKS_PER_SEC;    
    cout << "done in " << seconds << " seconds." << endl;
    cout << "Average " << (seconds / counter) << " seconds/get" << endl << endl;;

    cout << "Begining slicing tests..." << endl;
    
    start = clock();
    for(int i = 0; i < cube.getNumberOfDimensions()-1; i++){
	int val =  (int)floor((double)rand()/RAND_MAX*(2<<(numBits/numDims)>>1));
//	cout << "Slicing dimension " << i << " on value " << val << endl;
//	::unlink("slicecube");
	MultidimensionalArray<int>* sliced= cube.slice(i, val,"slicecube");
//	cout << "NumberOfRecords = " << sliced->getNumberOfRecords() << endl;
	sliced->deleteOnDisk();
	delete sliced;

	//Sanity Check
	iter = cube.iterator();
	count = 0;
	while(iter->next()){
	    vector<int> key = iter->key();
	    if(key[i] == val){
		count ++;
	    }
	}
	cout << "Overall found " << count << " in iterator." << endl;
	iter = cube.sliceIterator(i,val);
	count = 0;
	while(iter->next()){
	    count ++;
	}
	cout << "Overall found " << count << " in sliceIterator." << endl;

    }
    finish = clock();
    seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
    cout << "done in " << seconds << " seconds." << endl;
    cout << "Average " << (seconds / cube.getNumberOfDimensions()) << " per slice" << endl << endl;;
    
    cout << "About to delete the cube...";
    cube.deleteOnDisk();
    cout << "done." << endl;
}

void simpletest(){
    int numDims = 2;
    
    cout << "Creating Linear Hash Cube... ";
    LinearHashCube<int> cube(numDims, "linearhashtest.bin");
    cout << "done."<< endl;
    
    cout << "Getting the name of the cube... " << cube.getName() << endl;
    
    cout << "Filling the cube with zeros... ";
    cube.fillWithZeros();
    cout << "done."<< endl;
    
    vector<int> key(2);
    key[0] = 101;
    key[1] = 5;
    cout <<"Testing the interleave function on "<<key[0]<<", "<<key[1]<<"...";
    int val = cube.interleave(key);
    cout << val << endl;;
    
    cout << "Trying a put... ";
    cube.put(key,1);
    cout << "done."<<endl;

    cout << "Testing the getNumberOfRecords function... ";
    int numRecs = cube.getNumberOfRecords();
    cout << numRecs << "." << endl;
    
    cout << "Trying a get... ";
    int x = cube.get(key);
    cout << "I got "<< x << endl;
    
    
    cout << "Trying a put over the previous put... ";
    cube.put(key,2);
    cout << "done."<<endl;
    
    cout << "Testing the getNumberOfRecords function (should still be 1)... ";
    numRecs = cube.getNumberOfRecords();
    cout << numRecs << "." << endl;
    
    cout << "Trying a get... ";
    x = cube.get(key);
    cout << "I got "<< x << endl;
    
    key[1] = 6;
    cout << "Trying another put in a different location... ";
    cube.put(key,3);
    cout << "done."<<endl;
    
    cout << "Trying a get... ";
    x = cube.get(key);
    cout << "I got "<< x << endl;
    
    cout << "Testing the getNumberOfRecords function (should be 2)... ";
    numRecs = cube.getNumberOfRecords();
    cout << numRecs << "." << endl;
    
    key[0] = 0;
    cout << "Trying another put in a different location... ";
    cube.put(key,5);
    cout << "done."<<endl;
    
    cout << "Trying undefinedAt " << key[0] << ", " << key[1]<<"... ";
    bool y = cube.undefinedAt(key);
    cout << "I got " << y << endl;
    
    key[0] = 10;
    cout << "Trying undefinedAt " << key[0] << ", " <<key[1]<< "... ";
    y = cube.undefinedAt(key);
    cout << "I got " << y << endl;
    
    key[0] = 101;
    key[1] = 5;
    cout << "Trying undefinedAt " << key[0] << ", " <<key[1]<< "... ";
    y = cube.undefinedAt(key);
    cout << "I got " << y << endl;
    
    vector<int> start(2);
    start[0] = 0;
    start[1] = 0;
    vector<int> end(2);
    end[0] = 250;
    end[1] = 7;

    cout <<"trying iterator..." << endl;
    MultidimensionalArray<int>::IteratorPtr iter(cube.iterator(start,end));
    while(iter->next()){
	cout << iter->value() << endl;
    }
    cout << "done." << endl;

    cout << "trying sliceIterator..."  << endl;
    iter= (cube.sliceIterator(1,6));
    while(iter->next()){
	cout << iter->value() << endl;
    }
    cout << "done." << endl;

    cout << "trying diceIterator..."  << endl;
    iter= (cube.diceIterator(1,266));
    while(iter->next()){
	cout << iter->value() << endl;
    }
    cout << "done." << endl;

    cube.deleteOnDisk();
}

int main() {
    simpletest();
}

