//Author : Steven Keith can be blamed for this...
#ifndef PSQLCUBE_H
#define PSQLCUBE_H

#include <pqxx/pqxx>
#include <pqxx/connection>
#include <pqxx/tablewriter>
#include <pqxx/transaction>
#include <pqxx/result>
#include <string>
#include <fstream>
#include <cassert>
#include <map>
#include <set>
#include <sstream>
#include <iostream>
#include <stdlib.h>
#include <limits.h>
#include <stdio.h>
#include <map>
#include <stack>

#include "fileutil.h"
#include "multidimensionalarray.h"
#include "hint.h"
#include "bulkwriter.h"

using namespace std;
using namespace PGSTD;
using namespace pqxx;

template<class DataType>
class PostgreSQLCubeIterator; 
template<class DataType>
class PostgreSQLCubeSliceIterator;
template<class DataType>
class PostgreSQLCubeDiceIterator;


template<class DataType>
class PostgreSQLCube:public MultidimensionalArray<DataType> {
    
 public:
    PostgreSQLCube( int NumberOfDimensions, string filename = "BTreeCubeTemp.bin") throw(std::ifstream::failure);

    //should finish loading anything in the queue
    //deleting everything right now while developing 
    virtual ~PostgreSQLCube(){}; 
    
    string getName() const;
    
    void open() throw(std::ifstream::failure);
    
    void close(); 
    
    bool deleteOnDisk();

    void sync();
    
    void fillWithZeros();

    inline void put(const vector<int>& pos, DataType value);
    inline void put(const int pos[], DataType value);
    
    void setLoading(int b);
    
    inline DataType get( const vector<int>& pos) const;
    inline DataType get( const int pos[]) const;
    
    inline bool undefinedAt( const vector<int>& pos) const;
    inline bool undefinedAt( const int pos[]) const;
    
    int getNumberOfRecords();
    
    void printData();
    
    uint64 getSizeOnDisk();

    

    typename MultidimensionalArray<DataType>::
	IteratorPtr iterator(const vector<int>& start, const vector<int>& end) const {
	return typename MultidimensionalArray<DataType>::IteratorPtr(new PostgreSQLCubeIterator<DataType>(*this,start,end));
    }
    
    typename MultidimensionalArray<DataType>::
	IteratorPtr iterator() const {
	return typename MultidimensionalArray<DataType>::IteratorPtr(new PostgreSQLCubeIterator<DataType>(*this));
    }
    
    typename MultidimensionalArray<DataType>::IteratorPtr sliceIterator( int dim, int slice_at) {
	return new PostgreSQLCubeSliceIterator<DataType>(*this,dim,slice_at);
    }
    
    typename MultidimensionalArray<DataType>::IteratorPtr diceIterator(  int dim, int stop_after) {
	return new PostgreSQLCubeDiceIterator<DataType>(*this, dim, stop_after);
    }


    
   typename MultidimensionalArray<DataType>::MultidimensionalArrayPtr slice( const uint dim, const uint slice_at,
					    const string fname,const int CUBETYPE=0, Hint h = Hint()) throw(std::out_of_range,std::ifstream::failure, std::bad_alloc);

   typename MultidimensionalArray<DataType>::MultidimensionalArrayPtr dice( const uint dim, const uint slice_at,
					    const string fname,const int CUBETYPE=0, Hint h = Hint()) throw(std::out_of_range,std::ifstream::failure, std::bad_alloc);
    
   typename MultidimensionalArray<DataType>::MultidimensionalArrayPtr  
       general_dice(const InvertedAgMap map[], 
                    const string fname, 
                    const int CUBETYPE=0, 
                    Hint h = Hint()) 
       throw(std::ifstream::failure, std::bad_alloc, std::invalid_argument);
   
   typename MultidimensionalArray<DataType>::MultidimensionalArrayPtr 
       full_rollup( const vector<int>& dimsToKill, 
                    const string fname,
                    const int CUBETYPE=0,
                    Hint h = Hint()) const  
       throw(std::out_of_range, std::ifstream::failure, 
             std::bad_alloc, std::invalid_argument);

    

 protected:
    
 private:
    virtual MultidimensionalArray<DataType>* createNew(vector<int> shape,string file, Hint hint = Hint()) const  throw(std::ifstream::failure); 
    
    string recursiveDice(string last, int dim, InvertedAgMap map);

    int mNumberOfDimensions;
    Hint mHint;
    string mFilename;
    
    string PREFIX;
    enum {verbose=0};
    
    bool active;
    int loading;
    lazyconnection C;
    BulkWriter<int> bw;
    
    
    friend class PostgreSQLCubeIterator<DataType>; 
    friend class PostgreSQLCubeSliceIterator<DataType>;
    friend class PostgreSQLCubeDiceIterator<DataType>;

};


template<class DataType>
PostgreSQLCube<DataType>::PostgreSQLCube( int NumberOfDimensions, string filename) throw(std::ifstream::failure) : MultidimensionalArray<DataType>(vector<int>(NumberOfDimensions,this->INVALID_SIZE)),C("dbname=lemur user=n7sh4"), bw(filename){
    
    this->PREFIX = "[psqlcube.h] ";
    this->mFilename = filename;
    this->mNumberOfDimensions = NumberOfDimensions;
    this->active = false;
    this->loading = 0;
    this->open();
}

template<class DataType>
void PostgreSQLCube<DataType>::open() throw(std::ifstream::failure){

    this->active = true;

    try{
	work T(this->C, "psqlcube.h -- open()");
	if(verbose) cout<<this->PREFIX<<"checking to see if table "<<this->mFilename<<" exists..."<<endl;
	T.exec(("SELECT COUNT(measure) FROM " + this->mFilename).c_str() );
	//shouldn't get here unless the table filename exists... 
	//if it exists do nothing.

	return;

	//throw std::ifstream::failure("This table already exists");
    }catch(const sql_error &e){
	if(verbose) cout<< this->PREFIX<<"Good! Table " << this->mFilename << " does not exist"<<endl;
    }

    if(verbose) cout<<this->PREFIX<<"Creating the table " << this->mFilename <<" with " << this->mNumberOfDimensions<<" dimensions..." << endl;
    try{
	stringstream dims;
	stringstream dims2;
	for(int i = 0; i < this->mNumberOfDimensions; i++){
	    dims<< "d" << i << " INTEGER, ";
	    dims2<< "d" << i << "=new.d"<< i;
	    if(i < this->mNumberOfDimensions-1) dims2 << " AND ";
	}
	{
	    work T(this->C, "psqlcube.h -- open()");
	    T.exec(("CREATE TABLE " + this->mFilename + "(" + dims.str() + "measure INTEGER);").c_str() );
	    T.commit();
	}
	{
	    work T(this->C, "psqlcube.h -- open()");
	    T.exec(("CREATE FUNCTION " + this->mFilename + "_function() RETURNS TRIGGER AS 'BEGIN DELETE FROM " + this->mFilename + " WHERE " + dims2.str() + "; RETURN new; END;'LANGUAGE 'plpgsql';").c_str());
	    T.commit();
	}
	{
	    work T(this->C, "psqlcube.h -- open()");
	    T.exec(("CREATE TRIGGER " + this->mFilename + "_trigger BEFORE INSERT ON " + this->mFilename + " FOR EACH ROW EXECUTE PROCEDURE " + this->mFilename + "_function();").c_str());
	    T.commit();
            }
	if(verbose) cout << " I've created a table, a function, and a trigger" << endl;

//	C.disconnect();
    }catch(const sql_error &e){
	cerr << PREFIX<<"SQL error: " << e.what() << endl
	     << PREFIX<<"Query was: '" << e.query() << "'" << endl;
	throw std::ifstream::failure("SQL Error creating table");
    }
}



template<class DataType>
MultidimensionalArray<DataType>* PostgreSQLCube<DataType>::createNew(vector<int> shape, string file, Hint h) const  throw(std::ifstream::failure){
   return new PostgreSQLCube<DataType>(shape.size(),file);
}


template<class DataType>
void PostgreSQLCube<DataType>::setLoading(int b){
  if(!(b == 0 || b == 1)) return;
  if(b == 1 && this->loading == 0){
        try{
            stringstream dims;
            for(int i = 0; i < this->mNumberOfDimensions; i++){
                dims<< "d" << i << " INTEGER, ";
            }
	    /* First make sure the loading table does not exist... it shouldn't
	       unless this cube was created before and not switched out of its
	       loading state or someting went horribly wrong! */
	    try{
	      work T(this->C, "psqlcube.h -- open()");
	      T.exec(("DROP TABLE " + this->mFilename + "_LOADING;").c_str());
	      T.commit();
	    }catch(const sql_error &e){; /*do nothing... just avoiding error*/}
            {
                work T(this->C, "psqlcube.h -- open()");
                T.exec(("CREATE TABLE " + this->mFilename + "_LOADING(" + dims.str() + "measure INTEGER, index INTEGER DEFAULT 0);").c_str() );
                T.exec(("INSERT INTO " + this->mFilename + "_LOADING(SELECT * FROM " + this->mFilename + ");").c_str() );
                T.commit();
            }
	    bw.setIndex(1); //adding extra just to be sure!
            this->deleteOnDisk();
            bw.getBulkWriter();
            
        }catch(const sql_error &e){
            cerr << PREFIX<<"SQL error: " << e.what() << endl
                 << PREFIX<<"Query was: '" << e.query() << "'" << endl;
            throw std::ifstream::failure("SQL Error creating table");
        }
    }
    if(b == 0 && this->loading == 1){
        bw.deleteBulkWriter();
        
        try{
	  
	  stringstream dims;
	  stringstream dims2;
	  stringstream dims3;
	  stringstream dims4;
	  for(int i = 0; i < this->mNumberOfDimensions; i++){
	    dims<< "d" << i << ", ";
	    dims2<< "d" << i ;
	    if(i < this->mNumberOfDimensions -1) dims2 << ", ";
	    dims4<< "d" << i << " INTEGER, ";
	    dims3<< "d" << i << "=new.d"<< i;
	    if(i < this->mNumberOfDimensions-1) dims3 << " AND ";
	  }
	  
	  
	  {
	    work T(this->C, "psqlcube.h -- open()");
	    T.exec(("CREATE TABLE " + this->mFilename + "(" + dims4.str() + "measure INTEGER);").c_str() );
	    T.commit();
	  }
	  {
	    work T(this->C, "psqlcube.h -- open()");
	    T.exec(("INSERT INTO " + this->mFilename + " (SELECT DISTINCT ON(" + dims2.str() 
		    + ") "+dims.str()+ "measure FROM " + this->mFilename 
		    + "_LOADING ORDER BY " + dims.str() + " index DESC) ").c_str());
	    
	    T.exec(("DELETE FROM " + this->mFilename + "_LOADING").c_str() );
	    
	    T.exec(("DROP TABLE " + this->mFilename + "_LOADING").c_str() );
	    T.commit();
	  }
	  {
	    work T(this->C, "psqlcube.h -- open()");
	    T.exec(("CREATE FUNCTION " + this->mFilename + "_function() RETURNS TRIGGER AS 'BEGIN DELETE FROM " + this->mFilename + " WHERE " + dims3.str() + "; RETURN new; END;'LANGUAGE 'plpgsql';").c_str());
	    T.commit();
	  }
	  {
	    work T(this->C, "psqlcube.h -- open()");
	    T.exec(("CREATE TRIGGER " + this->mFilename + "_trigger BEFORE INSERT ON " + this->mFilename + " FOR EACH ROW EXECUTE PROCEDURE " + this->mFilename + "_function();").c_str());
	    T.commit();
	  }
	  
        }catch(const sql_error &e){
	  cerr << PREFIX<<"SQL error: " << e.what() << endl
	       << PREFIX<<"Query was: '" << e.query() << "'" << endl;
	  throw std::ifstream::failure("SQL Error creating table");
        }
        
    }
    this->loading = b;
    
}



template<class DataType>
string PostgreSQLCube<DataType>::getName() const{
  return this->mFilename;
}

template<class DataType>
void PostgreSQLCube<DataType>::sync(){
  return;
}

template<class DataType>
void PostgreSQLCube<DataType>::close(){
  if(this->active){
    this->active = false;
    this->setLoading(0);
    this->bw.deleteBulkWriter();
  }
  
}
template<class DataType>
bool PostgreSQLCube<DataType>::deleteOnDisk(){
  this->close();
  if(verbose) cout<<this->PREFIX<<"Dropping the table " << this->mFilename << endl;
  try{
    
    work T(this->C, "psqlcube.h -- close()");
    T.exec(("DROP TABLE " + this->mFilename + ";").c_str() );
    T.exec(("DROP FUNCTION " + this->mFilename + "_function() CASCADE;").c_str() );
    T.commit();
    //	C.disconnect();
  }catch(const sql_error &e){
    cerr << PREFIX<<"SQL error: " << e.what() << endl
	 << PREFIX<<"Query was: '" << e.query() << "'" << endl;
    throw std::ifstream::failure("SQL Error Dropping Table");
  }
  return true;
}

template<class DataType>
void PostgreSQLCube<DataType>::fillWithZeros(){
//    connection C("dbname=lemur user=steve");

    if(verbose) cout<<this->PREFIX<<"Deleting all values from " << this->mFilename << endl;
    try{
	work T(this->C, "psqlcube.h -- fillWithZeros()");
    	T.exec(("DELETE FROM " + this->mFilename + ";").c_str() );
	T.commit();
//	C.disconnect();
    }catch(const sql_error &e){
	cerr << PREFIX<<"SQL error: " << e.what() << endl
	     << PREFIX<<"Query was: '" << e.query() << "'" << endl;
    }
}

template<class DataType>
void PostgreSQLCube<DataType>::put(const vector<int>& pos, DataType value){
  if(loading){
	bw.insert(pos,value,this->mNumberOfDimensions);
    }else{
	stringstream dims;
	dims << "INSERT INTO " << this->mFilename <<"(";
	for(int i = 0; i < this->mNumberOfDimensions; i++){
	    dims << "d" << i << ", ";
	}
	dims << "measure) VALUES (";
	for(int i = 0; i < this->mNumberOfDimensions; i++){
	    dims << pos[i] << ", ";
	}
	dims << value << ");";
//	connection C("dbname=lemur user=steve");
	
	if(verbose) cout<<this->PREFIX<<"INSERTING the following statement into table " << this->mFilename << endl;
	if(verbose) cout<<this->PREFIX<<dims.str()<<endl;
	try{
	    work T(this->C, "psqlcube.h -- put()");
	    T.exec((dims.str()).c_str() );
	    T.commit();
//	    C.disconnect();
	}catch(const sql_error &e){
	    cerr << PREFIX<<"SQL error: " << e.what() << endl
		 << PREFIX<<"Query was: '" << e.query() << "'" << endl;
	}
	
    }

}

template<class DataType>
void PostgreSQLCube<DataType>::put(const int pos[], DataType value){
    if(loading){
	bw.insert(pos,value,this->mNumberOfDimensions);
    }else{
	stringstream dims;
	dims << "INSERT INTO " << this->mFilename <<"(";
	for(int i = 0; i < this->mNumberOfDimensions; i++){
	    dims << "d" << i << ", ";
	}
	dims << "measure) VALUES (";
	for(int i = 0; i < this->mNumberOfDimensions; i++){
	    dims << pos[i] << ", ";
	}
	dims << value << ");";
//	connection C("dbname=lemur user=steve");
	
	if(verbose) cout<<this->PREFIX<<"INSERTING the following statement into table " << this->mFilename << endl;
	if(verbose) cout<<this->PREFIX<<dims.str()<<endl;
	try{
	    work T(this->C, "psqlcube.h -- put()");
	    T.exec((dims.str()).c_str() );
	    T.commit();
//	    C.disconnect();
	}catch(const sql_error &e){
	    cerr << PREFIX<<"SQL error: " << e.what() << endl
		 << PREFIX<<"Query was: '" << e.query() << "'" << endl;
	}
	
    }
}

template<class DataType>
DataType PostgreSQLCube<DataType>::get(const vector<int>& pos) const{
//    if(this->bw.getNum() > 0) this->bw.commit();

    stringstream dims;
    dims << "SELECT measure FROM " << this->mFilename;
    dims << " WHERE ";
    for(int i = 0; i < this->mNumberOfDimensions - 1; i++){
	dims << "d" << i << "="<<pos[i] << " AND ";
    }
    dims << "d" <<(this->mNumberOfDimensions-1)<< "=" << pos[this->mNumberOfDimensions-1]<<";";
//    connection C("dbname=lemur user=n7sh4");
    if(verbose) cout<<this->PREFIX<<"Getting the value at the following position in the table " << this->mFilename << endl;
    if(verbose) cout<<this->PREFIX<<dims.str()<<endl;
    try{
        connection C2("dbname=lemur user=n7sh4");

	work T(C2, "psqlcube.h -- get()");
	result R = T.exec((dims.str()).c_str() );
	T.commit();
	C2.disconnect();
	if(R.size() == 0) return 0;
	stringstream sval;
	sval<< R[0][0];
	DataType val;
	sval >> val;
	return val;

    }catch(const sql_error &e){
	cerr << PREFIX<<"SQL error: " << e.what() << endl
	     << PREFIX<<"Query was: '" << e.query() << "'" << endl;
	return 0;
}
    
}

template<class DataType>
DataType PostgreSQLCube<DataType>::get(const int pos[]) const{
//    if(this->bw.getNum() > 0) this->bw.commit();

    stringstream dims;
    dims << "SELECT measure FROM " << this->mFilename;
    dims << " WHERE ";
    for(int i = 0; i < this->mNumberOfDimensions - 1; i++){
	dims << "d" << i << "="<<pos[i] << " AND ";
    }
    dims << "d" <<(this->mNumberOfDimensions-1)<< "=" << pos[this->mNumberOfDimensions-1]<<";";
//    connection C("dbname=lemur user=n7sh4");
    if(verbose) cout<<this->PREFIX<<"Getting the value at the following position in the table " << this->mFilename << endl;
    if(verbose) cout<<this->PREFIX<<dims.str()<<endl;
    try{
        connection C2("dbname=lemur user=n7sh4");
	work T(C2, "psqlcube.h -- get()");
	result R = T.exec((dims.str()).c_str() );
	T.commit();
	C2.disconnect();
	if(R.size() == 0) return 0;
	stringstream sval;
	sval<< R[0][0];
	DataType val;
	sval >> val;
	return val;

    }catch(const sql_error &e){
      cerr << PREFIX<<"SQL error: " << e.what() << endl
	   << PREFIX<<"Query was: '" << e.query() << "'" << endl;
      return 0;
    }
    
}


//The undefinedAt functions are INCORRECT... they do not actually work the way
//I know Owen wants them to...
template<class DataType>
bool PostgreSQLCube<DataType>::undefinedAt( const vector<int>& pos) const{

if(this->get(pos) == 0)
	return true;
    else
	return false;
    
}

template<class DataType>
bool PostgreSQLCube<DataType>::undefinedAt( const int pos[]) const{
  if(this->get(pos) == 0)
    return true;
  else
    return false;
  
}

template<class DataType>
uint64 PostgreSQLCube<DataType>::getSizeOnDisk(){
  if(this->bw.getNum() > 0) this->bw.commit();
  
  try{
    //      connection C2("dbname=lemur user=n7sh4");
    
    work T(this->C, "psqlcube.h -- getSizeOnDisk()");
    if(verbose) cout<<this->PREFIX<<"Getting the size of "<<this->mFilename<<endl;
    T.exec(("analyze " + this->mFilename).c_str());
    result R = T.exec(("SELECT relpages * 8192 AS size FROM pg_class WHERE relname = '" + this->mFilename + "'").c_str() );
    T.commit();
    //      C2.disconnect();
    if(R.size() == 0) return 0;
    stringstream sval;
    sval<< R[0][0];
    uint64 val;
    sval >> val;
    return val;
    
  }catch(const sql_error &e){
    cerr << PREFIX<<"SQL error: " << e.what() << endl
	 << PREFIX<<"Query was: '" << e.query() << "'" << endl;
    return 0;
  }
}

template<class DataType>
int PostgreSQLCube<DataType>::getNumberOfRecords(){
  //    connection C("dbname=lemur user=steve");
  
  if(this->bw.getNum() > 0) this->bw.commit();
  
  try{
    work T(this->C, "psqlcube.h -- getNUmberOfRecords()");
    if(verbose) cout<<this->PREFIX<<"Getting the number of records in table  "<<this->mFilename<<" exists..."<<endl;
    result R = T.exec(("SELECT COUNT(measure) FROM " + this->mFilename).c_str() );
    T.commit();
    //	C.disconnect();
    
    if(R.size() == 0) return 0;
    stringstream sval;
    sval<< R[0][0];
    int val;
    sval >> val;
    return val;
    
  }catch(const sql_error &e){
    cerr << PREFIX<<"SQL error: " << e.what() << endl
	 << PREFIX<<"Query was: '" << e.query() << "'" << endl;
    return 0;
  }
  
}

template<class DataType>
void PostgreSQLCube<DataType>::printData(){
    std::auto_ptr<PostgreSQLCubeIterator<DataType> > it(new PostgreSQLCubeIterator<DataType>(*this));
    
    while(it->next()){
	vector<int> key = it->key();
	for(uint i = 0; i < key.size(); i++){
	    cout<< key[i] << " ";
	}
	cout << it->value() << endl;
    }
    
}
template<class DataType>
typename MultidimensionalArray<DataType>::MultidimensionalArrayPtr PostgreSQLCube<DataType>::slice( const uint dim, const uint slice_at, const string fname ,const int CUBETYPE, Hint h) 
     throw(std::out_of_range,std::ifstream::failure, std::bad_alloc){
  MultidimensionalArray<DataType> * slicecube = new PostgreSQLCube<DataType>(this->mNumberOfDimensions-1, fname);
      try{
	work T(this->C, "psqlcube.h -- slice()");
	if(verbose) cout<<this->PREFIX<<"Slicing table  "<<this->mFilename<<" on dim "<< dim << " at " << slice_at<< endl;
	stringstream str;
	str << "INSERT INTO "<< fname << "(SELECT ";
	for(int i = 0; i < this->mNumberOfDimensions; i++){
	  if(i < (int)dim){
	    str << "d" << i << " as d" << i << ", " ;
	  }
	  if(i > (int)dim){
	    str << "d" << i << " as d" << (i-1) << ", " ;
	  }
	}
	str << "measure FROM " << this->mFilename << " WHERE d" << dim << " = " << slice_at << ")";

	if(verbose) cout << str.str();
	stringstream dims2;
	for(int i = 0; i < this->mNumberOfDimensions; i++){
	  dims2<< "d" << i << "=new.d"<< i;
	  if(i < this->mNumberOfDimensions-1) dims2 << " AND ";
	}

	T.exec(("DROP FUNCTION " + fname + "_function() CASCADE").c_str());
	T.exec((str.str()).c_str());
	T.exec(("CREATE FUNCTION " + fname 
		+ "_function() RETURNS TRIGGER AS 'BEGIN DELETE FROM " + fname 
		+ " WHERE " + dims2.str() + "; RETURN new; END;'LANGUAGE 'plpgsql';").c_str());
	T.exec(("CREATE TRIGGER " + fname + "_trigger BEFORE INSERT ON " + fname 
		+ " FOR EACH ROW EXECUTE PROCEDURE " + this->mFilename 
		+ "_function();").c_str());
      

	T.commit();

	return slicecube;
//	C.disconnect();
	
    }catch(const sql_error &e){
	cerr << PREFIX<<"SQL error: " << e.what() << endl
	     << PREFIX<<"Query was: '" << e.query() << "'" << endl;
	return 0;
    }

}
template<class DataType>
typename MultidimensionalArray<DataType>::MultidimensionalArrayPtr PostgreSQLCube<DataType>::dice( const uint dim, const uint slice_at, const string fname ,const int CUBETYPE, Hint h) 
     throw(std::out_of_range,std::ifstream::failure, std::bad_alloc){
  MultidimensionalArray<DataType> * dicecube = new PostgreSQLCube<DataType>(this->mNumberOfDimensions, fname);
      try{
	work T(this->C, "psqlcube.h -- slice()");
	if(verbose) cout<<this->PREFIX<<"Dicing table  "<<this->mFilename<<" on dim "<< dim << " at " << slice_at<< endl;
	stringstream str;
	str << "INSERT INTO "<< fname << "(SELECT ";

	for(int i = 0; i < this->mNumberOfDimensions; i++){
	  if(i != (int)dim){
	    str << "d" << i << " as d" << i << ", " ;
	  }
	  else{
	    str << "0 as d" << i << ", ";
	  }

	}


	str << "measure FROM " << this->mFilename << " WHERE d" << dim << " = " << slice_at << ")";

	stringstream dims2;
	for(int i = 0; i < this->mNumberOfDimensions; i++){
	  dims2<< "d" << i << "=new.d"<< i;
	  if(i < this->mNumberOfDimensions-1) dims2 << " AND ";
	}

	if(verbose) cout << str.str();

	T.exec(("DROP FUNCTION " + fname + "_function() CASCADE").c_str());
	T.exec((str.str()).c_str());
	T.exec(("CREATE FUNCTION " + fname 
		+ "_function() RETURNS TRIGGER AS 'BEGIN DELETE FROM " + fname 
		+ " WHERE " + dims2.str() + "; RETURN new; END;'LANGUAGE 'plpgsql';").c_str());
	T.exec(("CREATE TRIGGER " + fname + "_trigger BEFORE INSERT ON " + fname 
		+ " FOR EACH ROW EXECUTE PROCEDURE " + this->mFilename 
		+ "_function();").c_str());
      
	T.commit();

	return dicecube;
//	C.disconnect();
	
    }catch(const sql_error &e){
	cerr << PREFIX<<"SQL error: " << e.what() << endl
	     << PREFIX<<"Query was: '" << e.query() << "'" << endl;
	return 0;
    }

}

//NOTE: General Dicing does NOT support WEIGHTS!!!
template<class DataType>
typename MultidimensionalArray<DataType>::MultidimensionalArrayPtr  
PostgreSQLCube<DataType>::general_dice(const InvertedAgMap map[], 
                                       const string fname,
                                       int cubetype,
                                       Hint h) 
throw(std::ifstream::failure, std::bad_alloc, std::invalid_argument){

    stringstream s;
    s << this->mFilename;
    string lastfile = s.str();
    for(int i = 0; i < this->mNumberOfDimensions; i++){
        stringstream str;
        str << recursiveDice(lastfile, i, map[i]);
        lastfile = str.str();
    }
    MultidimensionalArray<DataType>* dicecube = new PostgreSQLCube(this->mNumberOfDimensions,
                                                                   fname);
    if(verbose)cout << "CUBE CREATED" << endl;
    
    try{

	stringstream dims2;
	for(int i = 0; i < this->mNumberOfDimensions; i++){
	    dims2<< "d" << i << "=new.d"<< i;
	    if(i < this->mNumberOfDimensions-1) dims2 << " AND ";
	}

      stringstream str;
      str << "INSERT INTO " << fname << "(SELECT * FROM "<< lastfile << ")";
      work T(this->C, "psqlcube.h -- general_dice()");
      T.exec(("DROP FUNCTION " + fname + "_function() CASCADE").c_str());
      T.exec((str.str()).c_str());
      T.exec(("CREATE FUNCTION " + fname 
	      + "_function() RETURNS TRIGGER AS 'BEGIN DELETE FROM " + fname 
	      + " WHERE " + dims2.str() + "; RETURN new; END;'LANGUAGE 'plpgsql';").c_str());
      T.exec(("CREATE TRIGGER " + fname + "_trigger BEFORE INSERT ON " + fname 
	      + " FOR EACH ROW EXECUTE PROCEDURE " + this->mFilename 
	      + "_function();").c_str());
      
      T.commit();
    }catch(const sql_error &e){
      cerr << PREFIX<<"SQL error: " << e.what() << endl
	   << PREFIX<<"Query was: '" << e.query() << "'" << endl;
      return 0;
    }
    if(lastfile.compare(this->mFilename) != 0){
      try{
        stringstream str;
        str << "DROP TABLE " << lastfile;
        work T(this->C, "psqlcube.h -- general_dice()");
        T.exec((str.str()).c_str());
        T.commit();
      }catch(const sql_error &e){
        cerr << PREFIX<<"SQL error: " << e.what() << endl
             << PREFIX<<"Query was: '" << e.query() << "'" << endl;
        return 0;
      }
    }
    return dicecube;
     
}

template<class DataType>
string PostgreSQLCube<DataType>::recursiveDice(string lastfile, int dim, InvertedAgMap map){
    if(map.isIdentity()) return lastfile;
    
    try{
	work T(this->C, "psqlcube.h -- recursiveDice()");
	stringstream str;
	str << "CREATE TABLE temp"<< dim << " ( ";
	for(int i = 0; i < this->mNumberOfDimensions; i++){
	    str << "d" << i << " INTEGER,";
	}
	str << "measure INTEGER)";
	if(verbose) cout << str.str();

	T.exec((str.str()).c_str());
	T.commit();
    }catch(const sql_error &e){
	cerr << PREFIX<<"SQL error: " << e.what() << endl
	     << PREFIX<<"Query was: '" << e.query() << "'" << endl;
	return 0;
    }

    for(int i = 0; i < map.mapSize(); i++){
        vector<int> ancestors = map.getAttributeAncestors(i);
        for(int j = 0; j < (int)ancestors.size(); j++){
	  if(verbose)cout << "i = "<< i << " j = " << j << endl;  
            try{
                stringstream str;
                str << "INSERT INTO temp" << dim << "(SELECT ";
                for(int z = 0; z < this->mNumberOfDimensions; z++){
                    if(z != dim){
                    str << "d" << z << " as d" << z << ", ";
                    }
                    else{
                        str << ancestors[j] << " as d" << z << ", "; 
                    }
                }
                str << "measure as measure from " << lastfile << " WHERE d" << dim << " = " << i << ")";
                work T(this->C, "psqlcube.h -- recursiveDice()");
                T.exec((str.str()).c_str());
                T.commit();
            
            }catch(const sql_error &e){
                cerr << PREFIX<<"SQL error: " << e.what() << endl
                     << PREFIX<<"Query was: '" << e.query() << "'" << endl;
                return 0;
            } //end try catch
        } //end for j
    } //end for i
    if(lastfile.compare(this->mFilename) != 0){
      try{
        stringstream str;
        str << "DROP TABLE "<< lastfile; 
        work T(this->C, "psqlcube.h -- recursiveDice()");
        T.exec((str.str()).c_str());
        T.commit();
      }catch(const sql_error &e){
        cerr << PREFIX<<"SQL error: " << e.what() << endl
             << PREFIX<<"Query was: '" << e.query() << "'" << endl;
        return 0;
      }
    }
    stringstream s;
    s << "temp" << dim;
    return s.str();
}


template <class DataType>
typename MultidimensionalArray<DataType>::MultidimensionalArrayPtr 
PostgreSQLCube<DataType>::full_rollup( const vector<int>& dimsToKill, 
                                       const string fname,
                                       const int CUBETYPE,
                                       Hint h) const  
throw(std::out_of_range, std::ifstream::failure, 
      std::bad_alloc, std::invalid_argument){
    MultidimensionalArray<DataType> * rollupcube = new PostgreSQLCube<DataType>(this->mNumberOfDimensions - (int)dimsToKill.size(), fname);
    try{
        connection C2("dbname=lemur user=n7sh4");
	work T(C2, "psqlcube.h -- full_rollup");
	if(verbose) cout<<this->PREFIX<<"Full_Rollup table  "<<this->mFilename<< endl;
	stack<int> s;
        stringstream str;
	str << "INSERT INTO "<< fname << "(SELECT ";
	stringstream dims;
        for(int i = 0; i < this->mNumberOfDimensions; i++){
            int found = 0;
            for(uint j = 0; j < dimsToKill.size(); j++){
                if(dimsToKill[j] == i){
                    found++;
                    break;
                }
            }
            if(found == 0){
                s.push(i);
            }
        }
        
        for(uint i = 0; i < s.size()-1; i++){
            dims << "d" << s.top() << ", ";
            s.pop();
        }
        dims << "d" << s.top();
        str << dims.str() << ", SUM(measure) FROM " << this->mFilename
            << " GROUP BY " << dims.str() << ")";
        
	if(verbose) cout << str.str();

	stringstream dims2;
	for(int i = 0; i < this->mNumberOfDimensions; i++){
	    dims2<< "d" << i << "=new.d"<< i;
	    if(i < this->mNumberOfDimensions-1) dims2 << " AND ";
	}


	T.exec(("DROP FUNCTION " + fname + "_function() CASCADE").c_str());
	T.exec((str.str()).c_str());
	T.exec(("CREATE FUNCTION " + fname 
		+ "_function() RETURNS TRIGGER AS 'BEGIN DELETE FROM " + fname 
		+ " WHERE " + dims2.str() + "; RETURN new; END;'LANGUAGE 'plpgsql';").c_str());
	T.exec(("CREATE TRIGGER " + fname + "_trigger BEFORE INSERT ON " + fname 
		+ " FOR EACH ROW EXECUTE PROCEDURE " + this->mFilename 
		+ "_function();").c_str());
      

	T.commit();
        C2.disconnect();
	return rollupcube;
    }catch(const sql_error &e){
	cerr << PREFIX<<"SQL error: " << e.what() << endl
	     << PREFIX<<"Query was: '" << e.query() << "'" << endl;
	return 0;
    }
    
    
    
}



template <class DataType>
class PostgreSQLCubeIterator:public MultidimensionalIterator<DataType> {

 public:
    PostgreSQLCubeIterator(const PostgreSQLCube<DataType> & owner,const vector<int>& start, const vector<int>& end) :
      MultidimensionalIterator<DataType>(),
	    index(-1),R(),
	    mStart(start), mEnd(end),mOwner(owner),  mKey(mStart), mCurrentValue(0), PREFIX ("[PostgreSQLCubeIterator] ")
{
	this->getValues();
	    }

    PostgreSQLCubeIterator(const PostgreSQLCube<DataType> & other):
      MultidimensionalIterator<DataType>(),
	    index(-1),R(),
	    mStart(other.getShape().size()),mEnd(other.getShape().size()), mOwner(other), mKey(other.getShape().size()),mCurrentValue(0),
	PREFIX ("[PostgreSQLCubeIterator] ")
{
	for(uint i = 0; i < mStart.size(); i++){
	    this->mStart[i] = 0;
	}
	for(uint i = 0; i < mEnd.size(); i++){
	    this->mEnd[i] = 99999999;
	}
	this->getValues();

    }

    PostgreSQLCubeIterator(const PostgreSQLCubeIterator<DataType> & other):
      MultidimensionalIterator<DataType>(),
      index(other.index),R(other.R),
      mStart(other.mStart), mEnd(other.mEnd), mOwner(other.mOwner), mKey(other.mKey), 
	mCurrentValue(other.mCurrentValue),  
	PREFIX( "[PostgreSQLCubeIterator] ")
{
}


    inline bool next(){
	index++;
	stringstream temp;
	temp << this->R.size();
	int x;
	temp >> x;
	if(index == x) return false;
	for(uint i = 0; i < this->mKey.size(); i++){
	    stringstream s;
	    s << this->R[index][i];
	    s >> this->mKey[i];
	}
	stringstream s;
	s << this->R[index][(int) mKey.size()];
	s >> this->mCurrentValue;
	return true;
    }

    inline DataType value() const{
	return this->mCurrentValue;
    }

    inline vector<int>& key(){
	return this->mKey;
    }
    
    PostgreSQLCubeIterator<DataType> * clone() const {return new PostgreSQLCubeIterator<DataType>(*this);}
 protected:
    int index;
    result R;
    vector<int> mStart, mEnd;
    const PostgreSQLCube<DataType> & mOwner;
    vector<int> mKey;
    DataType mCurrentValue;
    string PREFIX;
    enum {verbose=0};

    void getValues(){
	connection C("dbname=lemur user=n7sh4");

	try{
	    work T(C, "psqlcube.h -- getValues()");
	    if(verbose) cout<<this->PREFIX<<"Getting all  records in table  "<<this->mOwner.getName()<<" for the iterator"<<endl;
	    stringstream dims;
	    dims <<"SELECT * FROM " << mOwner.getName() << " WHERE ";
	    for(uint i = 0; i < this->mStart.size(); i++){
		dims << "d" << i << ">=" << this->mStart[i] << " AND ";
	    } 
	    for(uint i = 0; i < this->mEnd.size()-1; i++){
		dims << "d" << i << "<=" << this->mEnd[i] << " AND ";
	    }

	    dims << "d" << (this->mEnd.size() - 1) << "<=" << this->mEnd[mEnd.size()-1] << ";";
	    this->R = T.exec((dims.str()).c_str() );

	    T.commit();
	    C.disconnect();
	    
	}catch(const sql_error &e){
	    cerr << PREFIX<<"SQL error: " << e.what() << endl
		 << PREFIX<<"Query was: '" << e.query() << "'" << endl;
	}
	
    }
    
};

template <class DataType>
class PostgreSQLCubeSliceIterator:public MultidimensionalIterator<DataType> {

 public:
    PostgreSQLCubeSliceIterator(const PostgreSQLCube<DataType> & owner,int dim, int at) :
      MultidimensionalIterator<DataType>(), 
	    index(-1),
	    mDim(dim),
	    mAt(at),R(),
	    mOwner(owner), mKey(owner.getShape().size()), mCurrentValue(0),
	PREFIX ( "[PostgreSQLCubeSliceIterator] ")
	    
    {
	this->getValues();
    }

    
    PostgreSQLCubeSliceIterator(const PostgreSQLCubeSliceIterator<DataType> & other):
	MultidimensionalIterator<DataType>(),
        index(other.index),
	mDim(other.mDim),
	    mAt(other.mAt), R(other.R),
	    mOwner(other.mOwner), mKey(other.mKey), 
	mCurrentValue(other.mCurrentValue),
	    PREFIX ( "[PostgreSQLCubeSliceIterator] ") {
    }
    
    
    inline bool next(){
	index++;
	stringstream temp;
	temp << this->R.size();
	int x;
	temp >> x;
	if(index == x) return false;
	for(uint i = 0; i < this->mKey.size(); i++){
	    stringstream s;
	    s << this->R[index][i];
	    s >> this->mKey[i];
	}
	stringstream s;
	s << this->R[index][(int) mKey.size()];
	s >> this->mCurrentValue;
	return true;
    }

    inline DataType value() const{
	return this->mCurrentValue;
    }

    inline vector<int>& key(){
	return this->mKey;
    }
    
    PostgreSQLCubeSliceIterator<DataType> * clone() const {return new PostgreSQLCubeSliceIterator<DataType>(*this);}
 protected:

    int index, mDim, mAt;
    result R;
    const PostgreSQLCube<DataType> & mOwner;
    vector<int> mKey;
    DataType mCurrentValue;
    string PREFIX;
    enum {verbose=0};

    void getValues(){
	connection C("dbname=lemur user=n7sh4");

	try{
	    work T(C, "psqlcube.h -- getValues()");
	    if(verbose) cout<<this->PREFIX<<"Getting all  records in table  "<<this->mOwner.getName()<<" for the iterator"<<endl;
	    stringstream dims;
	    dims <<"SELECT * FROM " << mOwner.getName() << " WHERE ";
	    dims << "d" << this->mDim << "=" << this->mAt << ";";
	    if(verbose) cout << this->PREFIX << dims.str()<<endl;
	    this->R = T.exec((dims.str()).c_str() );

	    T.commit();
	    C.disconnect();
	    
	}catch(const sql_error &e){
	    cerr << PREFIX<<"SQL error: " << e.what() << endl
		 << PREFIX<<"Query was: '" << e.query() << "'" << endl;
	}
	
    }
    
};

template <class DataType>
class PostgreSQLCubeDiceIterator:public MultidimensionalIterator<DataType> {

 public:
    PostgreSQLCubeDiceIterator(const PostgreSQLCube<DataType> & owner,int dim, int stop_at) :
      MultidimensionalIterator<DataType>(), index(-1), 
	    mDim(dim),mStop_At(stop_at), R(), 
	    mOwner(owner), mKey(owner.getShape().size()), 
	    mCurrentValue(0),PREFIX ("[PostgreSQLCubeDiceIterator] "){
	this->getValues();
    }

    
    PostgreSQLCubeDiceIterator(
		    const PostgreSQLCubeDiceIterator<DataType> & other):
	MultidimensionalIterator<DataType>(),
	index(other.index),mDim(other.mDim),mStop_At(other.mStop_At),
	R(other.R),mOwner(other.mOwner),  mKey(other.mKey),
	mCurrentValue(other.mCurrentValue),
    	PREFIX ("[PostgreSQLCubeDiceIterator] ") {
    }
    
    
    inline bool next(){
	index++;
	stringstream temp;
	temp << this->R.size();
	int x;
	temp >> x;
	if(index == x) return false;
	for(uint i = 0; i < this->mKey.size(); i++){
	    stringstream s;
	    s << this->R[index][i];
	    s >> this->mKey[i];
	}
	stringstream s;
	s << this->R[index][(int) mKey.size()];
	s >> this->mCurrentValue;
	return true;
    }

    inline DataType value() const{
	return this->mCurrentValue;
    }

    inline vector<int>& key(){
	return this->mKey;
    }
    
    PostgreSQLCubeDiceIterator<DataType> * clone() const {return new PostgreSQLCubeDiceIterator<DataType>(*this);}
 protected:

    int index, mDim, mStop_At;
    result R;
    const PostgreSQLCube<DataType> & mOwner;
    vector<int> mKey;
    DataType mCurrentValue;
    string PREFIX;
    enum {verbose=0};

    void getValues(){
	connection C("dbname=lemur user=n7sh4");

	try{
	    work T(C, "psqlcube.h -- getValues()");
	    if(verbose) cout<<this->PREFIX<<"Getting all  records in table  "<<this->mOwner.getName()<<" for the iterator"<<endl;
	    stringstream dims;
	    dims <<"SELECT * FROM " << mOwner.getName() << " WHERE ";
	    dims << "d" << this->mDim << "<" << this->mStop_At << ";";
	    if(verbose) cout << this->PREFIX << dims.str()<<endl;
	    this->R = T.exec((dims.str()).c_str() );

	    T.commit();
	    C.disconnect();
	    
	}catch(const sql_error &e){
	    cerr << PREFIX<<"SQL error: " << e.what() << endl
		 << PREFIX<<"Query was: '" << e.query() << "'" << endl;
	}
	
    }
    
};

#endif
