#include "bigarray.h"
#include "internalbigarray.h"
#include "btreecube.h"
#include "btreecube_rotated.h"
#include "cdbcube.h"
#include <time.h>
#include <cmath>
#include <stdlib.h>


/**
* This file serves to test the speed of rotating.
* Steven Keith, November 4, 2005.
**/

void rotationTest(BTreeCube_Rotated<int>& bt, int d, vector<int> maxd, int records, int seed) {
   clock_t start,finish;
    double seconds;;
    int verbose = 0;
    cout << "Building a cube with "<<records<<" records in it, and "<< d <<" dimensions.\\\\"<<endl;
    cout << "Content of each dimension will have maximum values as follows... \\\\" << endl;
    for(int i = 0; i < d; i++){
	cout << "Dimension " << i << " has max value " << maxd[i] <<"\\\\"<< endl;
    }
//    cout << "Filling the cube with random content..." << endl;
    // fill the file with random contant
    vector<int> pos(d);
    srand(seed);// this, hopefully, makes the random generator deterministic (same numbers generated from run to run)
    start = clock();
    for (int k = 0; k < records; ++k) {
	int times = 0;
	//cout<< "entering record " << k << endl;
	while(true) {
	    for (int dim = 0; dim < d;++ dim)
		pos[dim] = (int) floor((double)rand()*maxd[dim]/RAND_MAX);
	    //cout << "got the location " << endl;
	    if(bt.get(pos) == 0) {
		bt.put(pos,1);
		break;
	    }
	    ++ times;
	    if(times > 2) {
		cout << "cube too dense, aborting\\\\"<<endl;
		break;
	    }
	}
	if(k %10000 == 0)  cout << k << " records written... ("<< 100.0*k/records <<"%)\\\\ "<< endl;
    }
    if(false) {
	cout <<" dumping the cube...\\\\ " <<endl;
	bt.printData();
	cout << " done dumping the cube...\\\\ " << endl;
    }
    finish = clock();
    seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
    cout << "Cube completed in "<< seconds << "seconds\\\\"<<endl;
    int numRecords = bt.getNumberOfRecords();
    cout << "Number of records : "<< numRecords <<"\\\\"<< endl;
    int numberOfTests = 4;   
    cout << "\\begin{tabular}{|c|c|c|c|c|c|}"<< endl;
    cout << "\\hline "<<endl;
    cout << "T$\\backslash$F & 1 & 2 & 3 & 4 & 5\\\\ \\hline" <<endl;
    for(int fromDim = 0; fromDim < d; ++fromDim){
	cout << fromDim << " ";
	for (int dim = 1; dim < d; ++dim) {
	    if(dim != fromDim){
		//make sure the rotation we are rotating from is actually rotated.
		bt.rotate(fromDim);
                if(verbose) cout << "checking rotation from "<< 0<< " to "<< fromDim<<endl;
//                bt.checkrotation(fromDim,0);
                if(verbose) cout << "checking rotation from "<< fromDim<< " to "<< 0<<endl;
//                bt.checkrotation(0,fromDim);
		
//		cout << endl << " =============================" << endl;
//		cout << "Rotating on dimension " << dim << " from dimension " << fromDim << endl;
		float totaltime = 0.0;
		for(int i = 0; i < numberOfTests; i++){
		    start = clock();
		    bt.rotate(dim,fromDim);
		    finish = clock();
		    seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
//		    cout << "Test " << (i+1) << " took " << seconds << " seconds." <<endl;
		    totaltime += seconds;
		    //Sanity checking... not necessary now that things work
                    if(verbose)cout << "checking rotation from "<< fromDim<< " to "<< dim<<endl;
//                    bt.checkrotation(dim,fromDim);
                    if(verbose)cout << "checking rotation from "<< dim<< " to "<< fromDim<<endl;
//                    bt.checkrotation(fromDim,dim);
//		    if(numRecords != bt.getNumberOfRecords(dim))
//			cout<<"Rotation is corrupt, has only "<<bt.getNumberOfRecords(dim)<<" records"<<endl;

		    bt.deleteRotation(dim);
		}
		
		cout << "& "<< totaltime/numberOfTests ;
	    }else{
		cout << "& -- ";
	    }

	}
	cout << "\\\\ \hline" << endl;
	bt.deleteRotation(fromDim);
    }

    cout<< "\\end{tabular}\\\\\\\\" << endl << endl;
}


int main(char **) {
        char * bigfile = "bigchunk.bin";

        int d = 6;
        ::unlink(bigfile);
        Hint h;
        BTreeCube_Rotated<int> btr(d,bigfile,h);

	for(int numRecords = 500000; numRecords < 1000000; numRecords *= 10){
	    for(int seed = 0; seed < 10; seed+=5){
		
		vector<int> maxd(d);
/*		maxd[0] = 1000;
		maxd[1] = 1000;
		maxd[2] = 1000;
		maxd[3] = 1000; 
		maxd[4] = 1000;
		maxd[5] = 1000;
		rotationTest(btr,d,maxd,numRecords, seed);
		btr.close();
		btr.deleteOnDisk();
		
		btr.open();*/
		maxd[0] = 10;
		maxd[1] = 100;
		maxd[2] = 1000;
		maxd[3] = 10000; 
		maxd[4] = 100000;
		maxd[5] = 1000000;
		rotationTest(btr,d,maxd,numRecords, seed);
		btr.close();
		btr.deleteOnDisk();
		
		btr.open();
		maxd[0] = 1000000;
		maxd[1] = 100000;
		maxd[2] = 10000;
		maxd[3] = 1000; 
		maxd[4] = 100;
		maxd[5] = 10;
		rotationTest(btr,d,maxd,numRecords, seed);
		btr.close();
		btr.deleteOnDisk();
		
		btr.open();
		maxd[0] = 100000;
		maxd[1] = 10;
		maxd[2] = 1000000;
		maxd[3] = 100; 
		maxd[4] = 10000;
		maxd[5] = 1000;
		rotationTest(btr,d,maxd,numRecords, seed);
		btr.close();
		btr.deleteOnDisk();
		btr.open();
	    }
	}
	btr.close();
}
