#include "bigarray.h"
#include "internalbigarray.h"
#include "btreecube.h"
#include "btreecube_rotated.h"
#include "cdbcube.h"
#include <time.h>
#include <cmath>
#include <stdlib.h>


/**
* This file serves to test the speed of rotating.
* Steven Keith, November 4, 2005.
**/

void rotationTest(BTreeCube_Rotated<int>& bt, int d, vector<int> maxd) {
   clock_t start,finish;
    double seconds;;
    vector<int> key(6);
    start = clock();
    for (int i = 0; i < maxd[0]; ++i) {
	key[0] = i;
	for(int j = 0; j < maxd[1]; j++){
	    key[1] = j;
	    for(int k = 0; k < maxd[2]; k++){
		key[2] = k;
		for(int l = 0; l < maxd[3]; l++){
		    key[3] = l;
		    for(int m = 0; m < maxd[4]; m++){
			key[4] = m;
			for(int n = 0; n < maxd[5]; n++){
			    key[5] = n;
			    bt.put(key,1);
			    
			}  
		    }  
		}  
	    }
	}
    }
    finish = clock();
    seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
    cout << "Cube completed in "<< seconds << "seconds\\\\"<<endl;
    int numRecords = bt.getNumberOfRecords();
    cout << "Number of records : "<< numRecords <<"\\\\"<< endl;
    int numberOfTests = 4;   
    cout << "\\begin{tabular}{|c|c|c|c|c|c|}"<< endl;
    cout << "\\hline "<<endl;
    cout << "T$\\backslash$F & 1 & 2 & 3 & 4 & 5\\\\ \\hline" <<endl;
    for(int fromDim = 0; fromDim < d; ++fromDim){
	cout << fromDim << " ";
	for (int dim = 1; dim < d; ++dim) {
	    if(dim != fromDim){
		//make sure the rotation we are rotating from is actually rotated.
		bt.rotate(fromDim);
                if(verbose) cout << "checking rotation from "<< 0<< " to "<< fromDim<<endl;
//                bt.checkrotation(fromDim,0);
                if(verbose) cout << "checking rotation from "<< fromDim<< " to "<< 0<<endl;
//                bt.checkrotation(0,fromDim);
		
//		cout << endl << " =============================" << endl;
//		cout << "Rotating on dimension " << dim << " from dimension " << fromDim << endl;
		float totaltime = 0.0;
		for(int i = 0; i < numberOfTests; i++){
		    start = clock();
		    bt.rotate(dim,fromDim);
		    finish = clock();
		    seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
//		    cout << "Test " << (i+1) << " took " << seconds << " seconds." <<endl;
		    totaltime += seconds;
		    //Sanity checking... not necessary now that things work
                    if(verbose)cout << "checking rotation from "<< fromDim<< " to "<< dim<<endl;
//                    bt.checkrotation(dim,fromDim);
                    if(verbose)cout << "checking rotation from "<< dim<< " to "<< fromDim<<endl;
//                    bt.checkrotation(fromDim,dim);
//		    if(numRecords != bt.getNumberOfRecords(dim))
//			cout<<"Rotation is corrupt, has only "<<bt.getNumberOfRecords(dim)<<" records"<<endl;

		    bt.deleteRotation(dim);
		}
		
		cout << "& "<< totaltime/numberOfTests ;
	    }else{
		cout << "& -- ";
	    }

	}
	cout << "\\\\ \hline" << endl;
	bt.deleteRotation(fromDim);
    }

    cout<< "\\end{tabular}\\\\\\\\" << endl << endl;
}


int main(char **) {
        char * bigfile = "bigchunk.bin";

        int d = 6;
        ::unlink(bigfile);
        Hint h;
        BTreeCube_Rotated<int> btr(d,bigfile,h);

	vector<int> maxd(d);
	maxd[0] = 6;
	maxd[1] = 6;
	maxd[2] = 6;
	maxd[3] = 6; 
	maxd[4] = 6;
	maxd[5] = 6;
	rotationTest(btr,d,maxd);
	btr.close();
	btr.deleteOnDisk();
	
	btr.open();
	maxd[0] = 2;
	maxd[1] = 4;
	maxd[2] = 6;
	maxd[3] = 8; 
	maxd[4] = 10;
	maxd[5] = 12;
	rotationTest(btr,d,maxd);
	btr.close();
	btr.deleteOnDisk();
	
	btr.open();
	maxd[0] = 12;
	maxd[1] = 10;
	maxd[2] = 8;
	maxd[3] = 6; 
	maxd[4] = 4;
	maxd[5] = 2;
	rotationTest(btr,d,maxd);
	btr.close();
	btr.deleteOnDisk();
	
	btr.open();
	maxd[0] = 10;
	maxd[1] = 2;
	maxd[2] = 12;
	maxd[3] = 4; 
	maxd[4] = 8;
	maxd[5] = 6;
	rotationTest(btr,d,maxd);
	btr.close();
	btr.deleteOnDisk();
}
