#include "bigarray.h"
#include "internalbigarray.h"
#include "btreecube.h"
#include "btreecube_rotated.h"
#include "cdbcube.h"
#include "psqlcube.h"
#include <time.h>
#include <cmath>
#include <stdlib.h>


/**
* This file serves to test the speed of slicing.
* Daniel Lemire, August 12th, 2005
**/

void slice(MultidimensionalArray<int>& bt) {
        int n = 1000;// cube is n*n*n
        int records = 4000;//1 million!
        clock_t start,finish;
        double seconds;
        int d = 3;
        cout << "Building a cube with "<<records<<" records in it, and "<< d <<" dimensions."<<endl;
        start = clock();
        cout << "Filling the cube with random content..." << endl;
        // fill the file with random contant
        vector<int> pos(d);
        srand(0);// this, hopefully, makes the random generator deterministic (same numbers generated from run to run)
        for (int k = 0; k < records; ++k) {
                int times = 0;
                //cout<< "entering record " << k << endl;
                while(true) {
                        for (int dim = 0; dim < d;++ dim)
                               pos[dim] = (int) floor((double)rand()*n/RAND_MAX);
                        //cout << "got the location " << endl;
                        if(bt.get(pos) == 0) {
                               bt.put(pos,1);
                               break;
                        }
                        ++ times;
                        if(times > 2) {
                                cout << "cube too dense, aborting"<<endl;
                                break;
                        }
                }
                if(k %10000 == 0)  cout << k << " records written... ("<< 100.0*k/records <<"%) "<< endl;
        }
        if(false) {
          cout <<" dumping the cube... " <<endl;
          bt.printData();
          cout << " done dumping the cube... " << endl;
        }
        finish = clock();
        seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
        cout << "Cube completed in "<< seconds << endl;
        cout << "Number of records : "<< bt.getNumberOfRecords() << endl;
        int numberoftests = 5;   
        bool verbose = false;
        for (int dim = 0; dim < d; ++dim) {
                cout << endl << " =============================" << endl;
                cout << " Trying to slice on dimension " << dim << endl;
                float totaltime = 0.0;
                for(int times = 0; times < numberoftests ; ++times) {
                        int val =  (int) floor((double)rand()*n/RAND_MAX);
                                ::unlink("slicecube");
                        start = clock();
                        std::auto_ptr<MultidimensionalIterator<int> >  it (bt.sliceIterator(dim,val)); 
                        while(it->next()) {
                          if(verbose) cout << " value " << it->value() << " at ";
                          vector<int> pos = it->key();
                          for(uint k = 0; k < pos.size();++k) {
                            if (verbose) cout << pos[k] << " " ;
                          }
                          if(verbose) cout << endl;
                          assert (pos[dim] == val);
                          //assert (! bt.undefinedAt(pos));
                        }
                        MultidimensionalArray<int>* sliced= bt.slice(dim, val,"slicecube");
                        cout << "slicing on attribute value : "<< val << " ";
			int slicedrecords = sliced->getNumberOfRecords();
                        cout << "number of records = " << slicedrecords<<endl;
                        if(verbose) sliced->printData();
                        sliced->deleteOnDisk();
                        delete sliced;
                        finish = clock();
                        seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
                        totaltime += seconds;
                        cout << "time = "<< seconds<< " s for "<< slicedrecords << " records" <<endl; 
                }
                cout << "average time is "<< totaltime/numberoftests<< " s"<< endl;
        }

}

void sqlslice(PostgreSQLCube<int>& bt) {
        
    int n = 1000;// cube is n*n*n
    int records = 4000;//1 million!
    clock_t start,finish;
    double seconds;
    int d = 3;
    bt.setLoading(1);
    cout << "Building a cube with "<<records<<" records in it, and "<< d <<" dimensions."<<endl;
    start = clock();
        cout << "Filling the cube with random content..." << endl;
        // fill the file with random contant
        vector<int> pos(d);
        srand(0);// this, hopefully, makes the random generator deterministic (same numbers generated from run to run)
        for (int k = 0; k < records; ++k) {
                int times = 0;
                //cout<< "entering record " << k << endl;
                while(true) {
                        for (int dim = 0; dim < d;++ dim)
                               pos[dim] = (int) floor((double)rand()*n/RAND_MAX);
                        //cout << "got the location " << endl;
			//                        if(bt.get(pos) == 0) {
                               bt.put(pos,1);
                               break;
			       //                        }
                        ++ times;
                        if(times > 2) {
                                cout << "cube too dense, aborting"<<endl;
                                break;
                        }
                }
                if(k %10000 == 0)  cout << k << " records written... ("<< 100.0*k/records <<"%) "<< endl;
        }
        if(false) {
          cout <<" dumping the cube... " <<endl;
          bt.printData();
          cout << " done dumping the cube... " << endl;
        }
	bt.setLoading(0);
        finish = clock();
        seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
        cout << "Cube completed in "<< seconds << endl;
        cout << "Number of records : "<< bt.getNumberOfRecords() << endl;
        int numberoftests = 5;   
        bool verbose = false;
        for (int dim = 0; dim < d; ++dim) {
                cout << endl << " =============================" << endl;
                cout << " Trying to slice on dimension " << dim << endl;
                float totaltime = 0.0;
                for(int times = 0; times < numberoftests ; ++times) {
                        int val =  (int) floor((double)rand()*n/RAND_MAX);
                                ::unlink("slicecube");
                        start = clock();
                        std::auto_ptr<MultidimensionalIterator<int> >  it (bt.sliceIterator(dim,val)); 
                        while(it->next()) {
                          if(verbose) cout << " value " << it->value() << " at ";
                          vector<int> pos = it->key();
                          for(uint k = 0; k < pos.size();++k) {
                            if (verbose) cout << pos[k] << " " ;
                          }
                          if(verbose) cout << endl;
                          assert (pos[dim] == val);
                          //assert (! bt.undefinedAt(pos));
                        }
                        MultidimensionalArray<int>* sliced= bt.slice(dim, val,"slicecube");
                        cout << "slicing on attribute value : "<< val << " ";
			int slicedrecords = sliced->getNumberOfRecords();
                        cout << "number of records = " << slicedrecords<<endl;
                        if(verbose) sliced->printData();
                        sliced->deleteOnDisk();
                        delete sliced;
                        finish = clock();
                        seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
                        totaltime += seconds;
                        cout << "time = "<< seconds<< " s for "<< slicedrecords << " records" <<endl; 
                }
                cout << "average time is "<< totaltime/numberoftests<< " s"<< endl;
        }

}


int main(char **) {
        char * bigfile = "bigchunk.bin";

        int d = 3;
        ::unlink(bigfile);
        BTreeCube<int> bt(d,bigfile);
        slice(bt);
        bt.close();
        bt.deleteOnDisk();
        Hint h;
//        for (int dim = 0; dim < d;++dim) h.addToFastSlicing(dim);
        for (int dim = 0; dim < d-1;++dim) h.addToFastSlicing(dim);
        BTreeCube_Rotated<int> btr(d,bigfile,h);
        slice(btr);
        btr.close();
        btr.deleteOnDisk();

	PostgreSQLCube<int> pc(d,"bigchunk");
	sqlslice(pc);
	pc.deleteOnDisk();
        pc.close();
	return 0;
}
