#include "btreecube.h"
#include "btreecube_rotated.h"
#include "bigarray.h"
#include "cdbcube.h"
#include "hint.h"

int main() {
  ::unlink("stevetest.bin");
  cout << "Here is a test." << endl;
  int numDims = 4;
  //vector<int> rotations(3);
  //rotations[0]=1; rotations[1]=2; rotations[2]=3;
  Hint hints;
  hints.addToFastSlicing(1);
  hints.addToFastSlicing(2);
  hints.addToFastSlicing(3);
  //hints.setFastSlicing(rotations);

  BTreeCube_Rotated<int> cube(numDims, "stevetest.bin", hints);
  cout << endl << "Beginning methods test..." << endl;
  cout << " getName() returns... " << cube.getName() << endl;

  cout << " putting 5 in slot [0,0,0,0] " << endl;
  vector<int> pos(4);
  pos[0] = 0;
  pos[1] = 0;
  pos[2] = 0;
  pos[3] = 0;
  int x = 5;
  cube.put(pos,x);

  cout << " putting 6 in slot [4,3,2,1] " << endl;
  pos[0] = 4;
  pos[1] = 3;
  pos[2] = 2;
  pos[3] = 1;
  x = 6;
  cube.put(pos,x);

  cout << " get(pos) returns... " << cube.get(pos) << endl;
  pos[0] = 0;
  pos[1] = 0;
  pos[2] = 0;
  pos[3] = 0;
  cout << " get([0,0,0,0]) returns... " << cube.get(pos) << endl;
  cout << " undefinedAt([0,0,0,0]) returns... " << cube.undefinedAt(pos) << endl;

  pos[3] = 1;
  cout << " undefinedAt([0,0,0,1]) returns... " << cube.undefinedAt(pos) << endl;


  cout << endl;
  cube.close();
  cube.deleteOnDisk();


  // follows some fun stuff
  ::unlink("dantest.bin");
  BTreeCube_Rotated<int> dcube(numDims, "dantest.bin", hints);
  for(int i = 0; i < 10;++i){
    vector<int>pos(numDims,i);
    dcube.put(pos,1);//diag matrix
  }
  for(int i = 0; i < 10;++i){
    vector<int>pos(numDims,i);
    if(dcube.get(pos)!= 1) cout << "bug!!!"<< endl;//diag matrix
  }
  // more fun stuff
  vector<int> bigshape(3);
  bigshape[0]=500;bigshape[1]=500;bigshape[2]=50;
  BTreeCube_Rotated<int> mbt(bigshape);
  mbt.fillWithZeros();  
  int xlength = 150;
  int ylength = 10;
  cout<<"[arrays] setting..." <<endl;
  int total = 0;
  int coordinates[3];

  coordinates[1] = coordinates[0] = 0; // local arrays are uninitialized in C++
  
  coordinates[2] = 3;
  mbt.put(coordinates,10);
  if(mbt.get(coordinates) !=10) { 
          cout << "bug!" << endl; return -2;
  }
  for(int x = 0; x < xlength; ++x) {
                for(int y = 0; y < ylength; ++y) {
                        for(int z= 0; z< 50; ++z) {
                                coordinates[0]=x;coordinates[1]=y;coordinates[2]=z;
                                mbt.put(coordinates,x+y+z);
                                ++total;
                                // immediate get to see if it works
                                if(mbt.get(coordinates)!= x+y+z) {
                                        cerr <<endl<<endl<<endl<<endl <<"BUG!!!! (immediate get) line " << __LINE__  <<endl;
                                        cerr << " I was expecting : "<< x+y+z<< endl;
                                        cerr << " I got : "<< mbt.get(coordinates)<< endl <<endl<<endl<<endl<<endl;
                                        cerr << " at: "<< x << " " << y << " " << z << endl;
                                        return -1;
                                 }
                        }
                }
  }
  //typename MultidimensionalArray<DataType>::IteratorPtr sliceIterator( int dim, int slice_at) const;   
  //typename MultidimensionalArray<DataType>::IteratorPtr diceIterator( int dim, int slice_at) const;   
  for(int x = 0; x < xlength; ++x) {
                //for(int y = 0; y < ylength; ++y) {
                                MultidimensionalIterator<int> *it = mbt.sliceIterator(0,x);
                                while(it->next()) {
                                  vector<int> coordinates = it->key();
                                  if(it->value() != coordinates[0]+coordinates[1]+coordinates[2]) {
                                                                cerr  <<endl<<endl<<endl<<endl<<"BUG!!!! line "<< __LINE__ <<endl <<endl<<endl<<endl<<endl;
                                  }
                                  if(coordinates[0] != x) {
                                                                cerr  <<endl<<endl<<endl<<endl<<"BUG!!!! line "<< __LINE__ <<endl <<endl<<endl<<endl<<endl;
                                  }
                                }
                                MultidimensionalIterator<int> *itd = mbt.diceIterator(0,x);
                                while(itd->next()) {
                                  vector<int> coordinates = itd->key();
                                  if(itd->value() != coordinates[0]+coordinates[1]+coordinates[2]) {
                                                                cerr  <<endl<<endl<<endl<<endl<<"BUG!!!! line "<< __LINE__ <<endl <<endl<<endl<<endl<<endl;
                                  }
                                  if(coordinates[0] > x) {
                                                                cerr  <<endl<<endl<<endl<<endl<<"BUG!!!! line "<< __LINE__ <<endl <<endl<<endl<<endl<<endl;
                                  }
                                }
                //}
  }
  for(int x = 0; x < ylength; ++x) {
                //for(int y = 0; y < ylength; ++y) {
                                MultidimensionalIterator<int> *it = mbt.sliceIterator(1,x);
                                while(it->next()) {
                                  vector<int> coordinates = it->key();
                                  if(it->value() != coordinates[0]+coordinates[1]+coordinates[2]) {
                                                                cerr  <<endl<<endl<<endl<<endl<<"BUG!!!! line "<< __LINE__ <<endl <<endl<<endl<<endl<<endl;
                                  }
                                  if(coordinates[1] != x) {
                                                                cerr  <<endl<<endl<<endl<<endl<<"BUG!!!! line "<< __LINE__ <<endl <<endl<<endl<<endl<<endl;
                                  }
                                }
                                MultidimensionalIterator<int> *itd = mbt.diceIterator(1,x);
                                while(itd->next()) {
                                  vector<int> coordinates = itd->key();
                                  if(itd->value() != coordinates[0]+coordinates[1]+coordinates[2]) {
                                                                cerr  <<endl<<endl<<endl<<endl<<"BUG!!!! line "<< __LINE__ <<endl <<endl<<endl<<endl<<endl;
                                  }
                                  if(coordinates[1] > x) {
                                                                cerr  <<endl<<endl<<endl<<endl<<"BUG!!!! line "<< __LINE__ <<endl <<endl<<endl<<endl<<endl;
                                  }
                                }
                //}
  }
  cout << " You passed. Don't forget to run 'make test; ./test' as well!" << endl;
  return 0;
}
