#ifndef STOREDLIST_H
#define STOREDLIST_H

#include <string>
#include <set>
#include <fstream>
#include <unistd.h>
#include <iostream>

using namespace std;

template <class T>
class StoredList {

 public:
  StoredList(string filename): mFileName(filename) {read();}

  ~StoredList() {}

  void add(T key) {mSet.insert(key); write();}

  void remove(T key) {mSet.erase(key); write();}

  set<T> get() {return mSet;}

  // returns 0 if ok
  int deleteOnDisk() {mSet.clear();return ::unlink(mFileName.c_str()) == 0;}
  bool isOnDisk() { return FileUtil::fileExists(mFileName.c_str()); }

  void clear() {mSet.clear(); write();}

  void dump() {
    for(typename set<T>::iterator it = mSet.begin(); it != mSet.end(); ++it) {
      cout << *it << endl;
    }
  }
  string getName() const { return mFileName;  }

 private:
  void read() {
    fstream f(mFileName.c_str(),ios_base::in);
    T key;
    f >> key;
    while( f ) {
      mSet.insert(key);
      f>>key;
    }
    f.close();
  }

  void write() {
    fstream f(mFileName.c_str(),ios_base::out | ios_base::trunc);
    if(!f) {
      cerr << "I didn't manage to open file " << mFileName.c_str()<<endl;
      return;
    }
    for(typename set<T>::iterator it = mSet.begin(); it != mSet.end(); ++it) {
      f << *it << endl;
    }
    if(f.fail()) {
      cerr << "Something didn't go right while writing to file " << mFileName.c_str()<<endl;
      return;
    }
    f.close();
  }

  string mFileName;
  set<T> mSet;
};
#endif
