#include "bigarray.h"
#include "internalbigarray.h"
#include "btreecube.h"
#include "btreecube_rotated.h"
#include "cdbcube.h"
#include <time.h>
#include <cmath>
#include <stdlib.h>


/**
* This file serves to test the speed of slicing.
* Daniel Lemire, August 12th, 2005
**/

void validate(BTreeCube_Rotated<int>& bt) {
        int n = 10000;// cube is n*n*n
        int records = 200000;//1 million!
        clock_t start,finish;
        double seconds;
        int d = 3;
        cout << "Building a cube with "<<records<<" records in it, and "<< d <<" dimensions."<<endl;
        start = clock();
        cout << "Filling the cube with random content..." << endl;
        // fill the file with random contant
        vector<int> pos(d);
        srand(0);// this, hopefully, makes the random generator deterministic (same numbers generated from run to run)
        for (int k = 0; k < records; ++k) {
                int times = 0;
                //cout<< "entering record " << k << endl;
                while(true) {
                        for (int dim = 0; dim < d;++ dim)
                               pos[dim] = (int) floor((double)rand()*n/RAND_MAX);
                        //cout << "got the location " << endl;
                        if(bt.get(pos) == 0) {
                               bt.put(pos,1);
                               break;
                        }
                        ++ times;
                        if(times > 2) {
                                cout << "cube too dense, aborting"<<endl;
                                break;
                        }
                }
                if(k %10000 == 0)  cout << k << " records written... ("<< 100.0*k/records <<"%) "<< endl;
        }
        if(false) {
          cout <<" dumping the cube... " <<endl;
          bt.printData();
          cout << " done dumping the cube... " << endl;
        }
        finish = clock();
        seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
        cout << "Cube completed in "<< seconds << "s" <<endl;
        cout << "Number of records : "<< bt.getNumberOfRecords() << endl;
        int numberoftests = 1;   
        bool verbose = false;
//	bt.printData();
        for (int dim = 1; dim < d; ++dim) {
	    cout << "Checking rotation "<<dim<< " from rotation 0"<<endl;
	    bt.checkrotation(dim,0);
	    cout << "Checking rotation "<<0<< " from rotation "<< dim<<endl;
	    bt.checkrotation(0,dim);
	}
	
}


int main(char **) {
        char * bigfile = "bigchunk.bin";

        int d = 3;
        ::unlink(bigfile);
        Hint h;
        for (int dim = 0; dim < d;++dim) h.addToFastSlicing(dim);
        BTreeCube_Rotated<int> btr(d,bigfile,h);
        validate(btr);
        btr.close();
        btr.deleteOnDisk();


}
